/******************************************************************************/
/*  %W% %D% %T%  */
/*                                                                            */
/*                                                                            */
/*                                                                            */
/*  Pascal-Programm:     Vera Kell                 Aachen, 31.07.1989         */
/*                                                                            */
/*  ANSI-C  Version:     Raphael M. Grochtmann     Aachen, 31.08.1990         */
/*                                                                            */
/*                                                                            */
/******************************************************************************/



#include <stdio.h>

#include "ext.h"
#include "list.h"
#include "levgen.h"



static void DelLevelVertex(plevel levlist, int *width)
   /* Loescht Ebenendummyknoten und alle damit verbundenen Kanten aus der Liste levlist. */
{
	int stateid, delnumber;
	pvertex qhelp;
	plevel levhelp;

	*width = 0;
	stateid = -1;
	levhelp = levlist;
	while (levhelp != NULL) {
		qhelp = levhelp->qlist;
		VertexListElemDel(&qhelp, stateid, &delnumber);
		if (levhelp->qnumber > *width)
		*width = levhelp->qnumber;
		levhelp->qlist = qhelp;
		LevelPositionSort(qhelp);
		levhelp = levhelp->levsucc;
	}
} /*DelLevelVertex*/



static void FindPriorities(boolean symm, plevel levliste, int qnr)
	/* Berechnung der Prioritaet eines Zustandes. Die Prioritaet richtet sich
	im Linksschritt nach der Anzahl der Vorgaenger(vert->barryleft),
	im Rechtsschritt nach der Anzahl der Nachfolger(vert->barryright).
	Dummyknoten, die einen Dummyknoten als Vorgaenger oder Nachfolger besitzen,
	erhalten die hoechste Prioritaet. */
{
	int qvornr, qnachnr;
	plevel levhelp;
	pvertex qlist;
	plist qpre, qsuc;

	levhelp = levliste;
	while (levhelp != NULL) {
		qlist = levhelp->qlist;
		while (qlist != NULL) {
			qvornr = 0;
			qpre = qlist->pre;
			while (qpre != NULL) {
				qvornr++;									/* Anzahl der Vorgaenger */
				if (qlist->q > qnr  &&  qpre->edge->q > qnr)
					qvornr = qnr+1;							/* Dummyknoten => hoechste Prioritaet */
				qpre = qpre->listsucc;
			}
			qlist->barryleft = qvornr;						/* Prioritaet von qlist->q (Linksschritt) */
			qnachnr = 0;
			qsuc = qlist->suc;
			while (qsuc != NULL) {
				qnachnr++;										/* Anzahl der Nachfolger */
				if (qlist->q > qnr  &&  qsuc->edge->q > qnr)
					qnachnr = qnr+1;							/* Dummyknoten => hoechste Prioritaet */
				qsuc = qsuc->listsucc;
			}
			qlist->barryright = qnachnr;					/* Prioritaet von qlist->q (Rechtsschritt) */
			if (symm)
				qlist->pos = 2*qlist->pos-1;
			qlist = qlist->vertsucc;
		}
		levhelp = levhelp->levsucc;
	}
} /*FindPriorities*/



static void PriorityLSort(pvertex qlist, matrixlevelarray levelmat, int *levcount)
	/* levelmat (und levcount) sind Var-Parameter */
	/* Sortieren der Zustaende einer Ebene in abfallender Reihenfolge anhand
	der Prioritaeten (Anzahl der Nachfolger), die in einem Linksschritt
	ermittelt worden sind. */
{
	int i, j;
	pvertex ql;

    *levcount = 0;											/* Anzahl der Zustaende der akt. Ebene */
    for (i = 1; i <= NMAX; levelmat[i++] = NULL)			/* levelmat: Zustaende der akt. Ebene */
		;
	while (qlist != NULL) {
		(*levcount)++;
		levelmat[*levcount] = qlist;
		qlist = qlist->vertsucc;
	}
	for (i = 1; i <= *levcount; i++)						/* Sortieren der Zustaende anhand von */
		for (j = 1; j <= *levcount - i; j++)				/* Prioritaeten (Anzahl der Vorgaenger) */
			if (levelmat[j+1]->barryleft  > levelmat[j]->barryleft
				||  levelmat[j+1]->barryleft == levelmat[j]->barryleft
				&&  levelmat[j+1]->pos < levelmat[j]->pos) {
				ql = levelmat[j+1];
				levelmat[j+1] = levelmat[j];
				levelmat[j] = ql;
			}
} /*PriorityLSort*/



static void PriorityRSort(pvertex qlist, matrixlevelarray levelmat, int *levcount)
	/* levelmat (und levcount) sind Var-Parameter */
	/* Sortieren der Zustaende einer Ebene in abfallender Reihenfolge anhand
	der Prioritaeten (Anzahl der Nachfolger), die in einem Rechtsschritt
	ermittelt worden sind. */
{
	int i, j;
	pvertex ql;

	*levcount = 0;												/* Anzahl der Zustaende der akt. Ebene */
	for (i = 1; i <= NMAX; levelmat[i++] = NULL)				/* levelmat: Zustaende der akt. Ebene */
		;
	while (qlist != NULL) {
		levelmat[++*levcount] = qlist;
		qlist = qlist->vertsucc;
	}
	for (i = 1; i <= *levcount; i++)							/* Sortieren der Zustaende anhand von */
		for (j = 1; j <= *levcount - i; j++)					/* Prioritaeten (Anzahl der Nachfolger) */
			if (levelmat[j + 1]->barryright > levelmat[j]->barryright
				||  levelmat[j + 1]->barryright == levelmat[j]->barryright
				&&  levelmat[j + 1]->pos < levelmat[j]->pos) {
				ql = levelmat[j + 1];
				levelmat[j + 1] = levelmat[j];
				levelmat[j] = ql;
			}
} /*PriorityRSort*/



static void ChangePosition(matrixposarray posorder, int chbegin, int chend, int chnr)
	/* posorder ist Var-Paramter */
{
	int j, mark;

	if (chbegin < chend) {
		mark = chbegin + 1;
		for (j = chbegin; j <= chend - chnr; j++) {
			while (posorder[mark].q == 0  &&  mark < chend)
				mark++;
			posorder[j].q = posorder[mark].q;
			posorder[j].vert = posorder[mark].vert;
			if (posorder[j].vert != NULL)
					posorder[j].vert->pos = j;
			posorder[mark].q = 0;
			posorder[mark].vert = NULL;
		}
		posorder[chend-chnr].exist = TRUE;
	} else {
		mark = chbegin - 1;
		for (j = chbegin; j >= chend + chnr; j--) {
			while (posorder[mark].q == 0  &&  mark > chend)
				mark--;
			posorder[j].q = posorder[mark].q;
			posorder[j].vert = posorder[mark].vert;
			if (posorder[j].vert != NULL)
				posorder[j].vert->pos = j;
			posorder[mark].q = 0;
			posorder[mark].vert = NULL;
		}
		posorder[chend + chnr].exist = TRUE;
	}
} /*ChangePosition*/



static void PriorityLPosition(boolean symm, pvertex levmatelem, int *position, int width)
	/* Berechnung der neuen Position eines Zustandes anhand des Mittelwertes
	der Vorgaenger. */
{
	plist levmatpre;
	double posreal;		/* wird nur bei symm gebraucht */;

	*position = 0;
	levmatpre = levmatelem->pre;
	while (levmatpre != NULL) {
		*position += levmatpre->edge->pos;
		levmatpre = levmatpre->listsucc;
	}
	if (levmatelem->barryleft <= width) {
		if (levmatelem->barryleft != 0) {
			if (symm)
				*position = round((double)*position / levmatelem->barryleft);
			else {
				posreal = (double)*position / levmatelem->barryleft;
				if (posreal - trunc(posreal) == 0.5) {
					if  (posreal < (width + 1) / 2.0)
						*position = round(posreal);
					else
						*position = trunc(posreal);
				} else 
					*position = round(posreal);
			}
		} else
			*position = 0;
	}
} /*PriorityLPosition*/



static void PriorityRPosition(boolean symm, pvertex levmatelem, int *position, int width)
	/* Berechnung der neuen Position eines Zustandes anhand des Mittelwertes
	der Nachfolger. */
{
	plist levmatsuc;
	double posreal; /* wird nur bei symm gebraucht */

	*position = 0;
	levmatsuc = levmatelem->suc;
	while (levmatsuc != NULL) {
		*position += levmatsuc->edge->pos;
		levmatsuc = levmatsuc->listsucc;
	}
	if (levmatelem->barryright <= width) {
		if (levmatelem->barryright != 0) {
			if (symm)
				*position = round((double)*position / levmatelem->barryright);
			else {
				posreal = (double)*position / levmatelem->barryright;
				if (posreal-trunc(posreal) == 0.5) {
					if (posreal < (width + 1) / 2.0)
						*position = round(posreal);
					else
						*position = trunc(posreal);
				} else
					*position = round(posreal);
			}
		} else
			*position = 0;
	}
} /*PriorityRPosition*/



static void PriorityPass(boolean symm, matrixlevelarray levmat, int levnr, int width, boolean leftpass)
	/* levmat ist Var-Parameter */
	/* Verschieben der Zustaende einer Ebene unter Beachtung der gegebenen
	Reihenfolge. */
{
	int j, jhelp, posold, posnew, posdist, posfree, posplace = 0; /* Zur Vorsicht initialisiert */
	matrixposarray posorder;
	pvertex distvert; /* wird nur bei symm gebraucht */

	for (j = 1; j <= NMAX; j++) {					/* Initialisierung der Positionsmatrix posorder */
		posorder[j].q = 0;
		posorder[j].exist = FALSE;
		posorder[j].vert = NULL;
	}
	if (symm)
		VertexListBeg(&distvert, -1, 0, 0, 0);
	for (j = 1; j <= levnr; j++) {					/* Zuordnen der Zustaende einer Ebene in posorder */
		posorder[levmat[j]->pos].q = levmat[j]->q;
		posorder[levmat[j]->pos].vert = levmat[j];
		if (symm) {
			posorder[levmat[j]->pos + 1].q = -1;
			posorder[levmat[j]->pos + 1].vert = distvert;
		}
	}
	for (j = 1; j <= levnr; j++) {
		posold = levmat[j]->pos;
		if (leftpass)
			PriorityLPosition(symm, levmat[j], &posnew, width);
		else
			PriorityRPosition(symm, levmat[j], &posnew, width);
		if (posnew == 0)
			posnew = posold;
		posdist = posnew - posold;
		if (posdist == 0)
			posorder[posold].exist = TRUE;
		else if (posdist < 0) {
			posfree = 0;
			jhelp = posold - 1;
			while (jhelp >= 1  &&  posorder[jhelp].exist != TRUE  &&  posfree < -posdist) {
				if (posorder[jhelp].q == 0) {
					posfree++;
					posplace = jhelp;
				}
				jhelp--;
			}
			if (posfree != 0)
				ChangePosition(posorder, posplace, posold, posfree);
			else
				posorder[j].exist = TRUE;
		} else if (posdist > 0) {
			posfree = 0;
			jhelp = posold + 1;
			while (posorder[jhelp].exist != TRUE  &&  posfree < posdist
				&&  jhelp <= (symm + 1) * width) {
				if (posorder[jhelp].q == 0) {
					posfree++;
					posplace = jhelp;
				}
				jhelp++;
			}
			if (posfree != 0)
				ChangePosition(posorder, posplace, posold, posfree);
			else
				posorder[j].exist = TRUE;
		}
	}
	if (symm)
		VertexListDel(&distvert);
} /*PriorityPass*/



void PriorityAlg(boolean symm, plevel levliste, int qnr, int width)
	/* Algorithmus zur Verfeinerung des Layouts anhand von Prioritaeten.
	Die Ebenenliste wird 3x (Links-, Rechts- ,Linksschritt) durchlaufen. */
{
	int					qnumber = 0, levnr; 	/* qnumber wird nur bei symm gebraucht */
												/* Zur Vorsicht initialisiert */
	plevel				levhelp, levl;
	matrixlevelarray	levhelpmat;

	DelLevelVertex(levliste, &width);
	FindPriorities(symm, levliste, qnr);		/* Prioritaeten (Anzahl der Vorgaenger) */
	levl = NULL;
	if (levliste != NULL)
		levhelp = levliste->levsucc;
	else
		levhelp = levliste;
	while (levhelp != NULL) {
		PriorityLSort(levhelp->qlist, levhelpmat, &levnr);
		PriorityPass(symm, levhelpmat, levnr, width, TRUE);
		levl = levhelp;
		levhelp = levhelp->levsucc;
	}
	levhelp = levl;
	while (levhelp != NULL) {
		PriorityRSort(levhelp->qlist, levhelpmat, &levnr);
		PriorityPass(symm, levhelpmat, levnr, width, FALSE);
		levl = levhelp;
		levhelp = levhelp->levprec;
	}
	if (symm) {
		levhelp = levliste;
		qnumber = 0;
	} else
		levhelp = levl;
	if (symm)
		while (levhelp != NULL  &&  levhelp->qnumber >= qnumber) {
			qnumber = levhelp->qnumber;
			levhelp = levhelp->levsucc;
		}
	else
		while (levhelp != NULL  &&  levhelp->qnumber != width)
			levhelp = levhelp->levsucc;
	while (levhelp != NULL) {
		PriorityLSort(levhelp->qlist, levhelpmat, &levnr);
		PriorityPass(symm, levhelpmat, levnr, width, TRUE);
		levhelp = levhelp->levsucc;
	}
} /*PriorityAlg*/
