/******************************************************************************/
/*  %W% %D% %T%  */
/*                                                                            */
/*                                                                            */
/*                                                                            */
/*  Pascal-Programm:     Vera Kell                 Aachen, 31.07.1989         */
/*                                                                            */
/*  ANSI-C  Version:     Raphael M. Grochtmann     Aachen, 31.07.1990         */
/*                                                                            */
/*  XWindows Version:    Raphael M. Grochtmann     Aachen, 31.08.1990         */
/*                                                                            */
/*                                                                            */
/******************************************************************************/

/*  Prozeduren zur graphischen Darstellung eines Automaten auf dem Bildschirm. 
	Bildschirm: Koordinatenursprung -> unten links
	Winkel -> 0 Grad liegt rechts,  Uhrzeigersinn  */


#include "ext.h"
#include "main.h"
#include "dfa.h"
#include "list.h"
#include "xgraphics.h"

#include <math.h>		/* Importiere: PI, sin(), cos(), atan(), sqrt() */
#define PI M_PI
#define arctan atan

#define sqr(a) (a)*(a)



static void FindGradient(double x, double y, double x1, double y1, double *gradient)
	/* Berechnung der Steigung einer Linie zwischen zwei Punkten (x, y) und (x1, y1) */
{
	if (x == x1)
		*gradient = 0;
	else
		*gradient = ((y1 - y) / (x1 - x));
} /* FindGradient */



void FindAngle(double x,  double y, double x1, double y1,  double *angle)
	/* Berechnung des Winkels zwischen X-Achse und Linie ((x, y), (x1, y1)).
	0 <= Winkel < 360 (Winkel=0,  falls y=y1 und x < x1) */
{
	double steigung, rad;

	FindGradient(x, y, x1, y1, &steigung);
	if (x == x1  &&  y == y1)
		*angle = 0;
	else {
		if (x == x1)
			if (y > y1) *angle = 270;
			else *angle = 90;
		if (y == y1)
			if (x < x1) *angle = 0;
			else *angle = 180;
		if (x != x1  &&  y != y1) {
			rad = arctan(steigung);
			*angle = rad * 180.0 / (double)PI;
			if (rad < 0) *angle += 180;
			if (y > y1)  *angle += 180;
		}
	}
} /* FindAngle */



static void AdaptAngle(double *angle)
	/* Winkel >= 360 => Winkel=Winkel-360;
	Winkel < 0    => Winkel=Winkel+360.  */
{
	int factor;

	factor = trunc(abs((int)*angle / 360));
	if (*angle >= 360)
		*angle -= factor * 360;
	if (*angle < 0)
		*angle += (factor + 1) * 360;
} /* AdaptAngle */



static void FindPoint(double *x, double *y, double angle, double length)
	/* Berechnung eines Punktes von (x, y) aus,  mit Hilfe der Richtung (Winkel)
	angle und dem Abstand length. Der neue Punkt wird durch (x, y) uebergeben. */
{
	double rad, deltax, deltay;

    rad = (angle * (double)PI) / 180.0;
	deltay = sin(rad) * length;
	*y += deltay;
	deltax = cos(rad) * length;
	*x += deltax;
} /* FindPoint */



static void FindLength(double x, double y, double x1, double y1, double *length)
	/* Berechnung des Abstandes zwischen zwei Punkten. */
{
	double deltax, deltay;

	deltax = abs((int)x - (int)x1);
	deltay = abs((int)y - (int)y1);
	deltax = sqr(deltax);
	deltay = sqr(deltay);
	*length = sqrt(deltax + deltay);
} /* FindLength */



static void FindTextLength(strg symbol, int *textlen)
	/* Diese Funktion liefert die dynamische Laenge eines strings zurueck. */
{
    *textlen = strlen(symbol) * TEXTWIDTH;
} /* FindTextLength */


static void FindTextBegPoint(double *x, double *y, double x1, double y1, strg symbol)
	/* Berechnung des Anfangspunktes fuer die Beschriftung der Transitionen (Kanten) */
{
	double	win, len, l;
	int		textwidth, intwin;

	FindAngle(*x, *y, x1, y1, &win);
	AdaptAngle(&win);
	FindLength(*x, *y, x1, y1, &len);
	FindTextLength(symbol, &textwidth);
	l = 1.0 / 3.0 * len;
	FindPoint(x, y, win, l);
	intwin = trunc(win);
	if (intwin == 0  ||  intwin  == 360) {
		*x -= (double)textwidth / 2;
		*y += FTBGOFFS;
	} else if (intwin == 180) {
		*x -= (double)textwidth / 2;
		*y -= TEXTHEIGHT + FTBGOFFS;
	} else if (intwin == 90) {
		*x -= textwidth + FTBGOFFS;
		*y -= (double)TEXTHEIGHT / 2.0;
	} else if (intwin == 270) {
		*x += FTBGOFFS;
		*y += (double)TEXTHEIGHT / 2.0;
	} else {
		if (0 < intwin  &&  intwin < 180)
			*x -= textwidth + FTBGOFFS;		/* (double)FTBGOFFS * 3.0 / 2.0;*/
		else
			*x += FTBGOFFS;					/* (double)FTBGOFFS * 3.0 / 2.0;*/
		if (90 < win  &&  win < 270)
			*y -= TEXTHEIGHT + FTBGOFFS;
		else
			*y += FTBGOFFS;
	}
} /* FindTextBegPoint */



void DrawSymbol(double x, double y, strg symbol)
	/* Beschriftung der Transitionen */
{
	int xint, yint, textlen, i, j;

	xint = round(x);
	yint = round(y);
	if (printer == FALSE) {
		if (SKAL_FAKTOR > 0.5)	/* wenn canvas zu klein, werden keine Symbole mehr gezeichnet */
			CText(xint, yint, symbol);
	} else {
		i = 2;
		while (symbol[i++] != '\0')
			;
		for (j = i; j <= STRLENGTH; j++)
			symbol[j] = ' ';	
		fprintf(druck_fp, "PU%i,%i;\n", xint, yint);
		fprintf(druck_fp, "LB%s%c\n", symbol, 3);		/* 3. ASCII-character ist CTRL-C */
	} 
} /* DrawSymbol */



static void DrawLine(double x, double y, double x1, double y1)
	/* Zeichne Linie zwischen zwei Punkten (x, y) und (x1, y1). */
{
	int xint, yint, x1int, y1int;

	xint = round(x);
	yint = round(y);
x1int = round(x1);
	y1int = round(y1);
	if (printer == FALSE)
		CLine(xint, yint, x1int, y1int);
	else {
	 	fprintf(druck_fp, "PU%i,%i;\n", xint, yint);
		fprintf(druck_fp, "PD%i,%i;\n", x1int, y1int);
	}
} /* DrawLine */



void DrawRedLine(double x,  double y,  double x1,  double y1,  double redlength,  strg sym)
{
	double winkel;

	FindAngle(x, y, x1, y1, &winkel);
	FindPoint(&x, &y, winkel, redlength);
	if (winkel >= 180)
		winkel -= 180;
	else
		winkel += 180;
	FindPoint(&x1, &y1, winkel, redlength);
	DrawLine(x, y, x1, y1);
	FindTextBegPoint(&x, &y, x1, y1, sym);
	DrawSymbol(x, y, sym);
} /* DrawRedLine */



static void DrawArrow (double x, double y, double x1, double y1, double arrowlength)
	/* Zeichne Pfeil zwischen zwei Punkten (x, y) und (x1, y1) in Richtung
	(x1, y1). Laenge der Pfeilspitze wird durch arrowlength angegeben. */
{
	double rad, winkel, mlen, alen, lwin, rwin, xm, ym, xl, yl, xr, yr;

	FindAngle(x1, y1, x, y, &winkel);
	rad = (20.0 * (double)PI) / 180.0;
	mlen = cos(rad) * arrowlength;
	xm = x1;
	ym = y1;
	FindPoint(&xm, &ym, winkel, mlen);
	alen = sin(rad) * arrowlength;
	lwin = winkel - 90;
	AdaptAngle(&lwin);
	xl = xm;
	yl = ym;
	FindPoint(&xl, &yl, lwin, alen);
	rwin = winkel + 90;
	AdaptAngle(&rwin);
	xr = xm;
	yr = ym;
	FindPoint(&xr, &yr, rwin, alen);
	DrawLine(x, y, x1, y1);
	DrawLine(x1, y1, xl, yl);
	DrawLine(x1, y1, xr, yr);
} /* DrawArrow */



void DrawRedArrow(double x,  double y,  double x1,  double y1, 
					double redlength,  double arrowlength,  strg sym)
	/* Zeichne Pfeil zwischen zwei Punkten (x, y) und (x1, y1),  wobei
	Pfeilanfang und Pfeilende den Abstand redlength von (x, y) und (x1, y1)
	besitzt. (Laenge der Pfeilspitze = arrowlength) */
{
	double winkel;

	FindAngle(x, y, x1, y1, &winkel);
	FindPoint(&x, &y, winkel, redlength);
	if (winkel >= 180)
		winkel -= 180;
	else
		winkel += 180;
	FindPoint(&x1, &y1, winkel, redlength);
	DrawArrow(x, y, x1, y1, arrowlength);
	FindTextBegPoint(&x, &y, x1, y1, sym);
	DrawSymbol(x, y, sym);
} /* DrawRedArrow */



void DrawDoubleRedArrow(double x,  double y,  double x1,  double y1, 
						double redlength,  double arrowlength,  strg sym)
	/* Zeichnet Pfeil (x, y) -> (x1, y1) parallel versetzt.
	Siehe ansonsten DrawRedArrow. (Laenge der Pfeilspitze = arrowlength) */
{
	double win, lwin;
	double x2, y2, x3, y3;

	FindAngle(x, y, x1, y1, &win);
	lwin = win + 90;
	if (lwin >= 360)
		lwin -= 360;
	x2 = x;
	y2 = y;
	x3 = x1;
	y3 = y1;
	FindPoint(&x, &y, lwin, DARROFFS);
	FindPoint(&x1, &y1, lwin, DARROFFS);
	DrawRedArrow(x, y, x1, y1, redlength, arrowlength, sym);
} /* DrawDoubleRedArrow */



void DrawBeginArrow(double x,  double y,  double redlength,  double arrowlength)
	/* Zeichne waagerechten Anfangspfeil (auf Startzustand). 2*arrowlength gibt
	die Laenge des Pfeiles an und length den Abstand der Pfeilspitze
	zum Startzustand. (Laenge der Pfeilspitze = arrowlength) */
{
	DrawArrow(x - (2 * arrowlength + redlength), y, x - redlength, y, arrowlength);
} /* DrawBeginArrow */



void DrawCircle(double x,  double y,  double radius)
	/* Zeichnet Kreis um den Mittelpunkt(x, y) mit dem Radius radius. */
{
	int xint, yint, radint;

	xint = round(x);yint = round(y);
	radint = round(radius);
	if (printer == FALSE)
		CCircle(xint, yint, radint);
	else {
		fprintf(druck_fp, "PU%i,%i;\n", xint, yint);
		fprintf(druck_fp, "CI%i;\n", radint);
	}
} /* DrawCircle */



void DrawState(double x,  double y,  double radius,  strg number)
	/* Zeichne Kreis um den Mittelpunkt (x, y) mit Radius radius und
	schreibe ein Zeichen number in diesen Kreis. */
{
	int		textwidth;
	double	deltax, deltay;

	DrawCircle(x, y, radius);
	if (strlen(number) > 2)	
		number[0] = '?';
	FindTextLength(number, &textwidth);
	deltay = (double)TEXTHEIGHT / 2.0;
	deltax = (double)textwidth / 2.0;
	DrawSymbol(x - deltax, y - deltay, number);
} /* DrawState */



static void DrawArc(double x, double y, double begwin, double endwin, double radius)
	/* Zeichnet Kreisbogen um den Mittelpunkt (x, y) mit Radius radius.
	Begwin legt den Anfangspunkt fest und endwin den Endpunkt. */
{
	int		xint, yint, xbeg, ybeg, xend, yend, radint;
	double	xbegr, ybegr, xendr, yendr;
	int		x1int, y1int, arcwin;

	xint = round(x);
	yint = round(y);
	if (printer == FALSE) {
		radint = trunc(radius);
		CArc(xint, yint, radint, begwin, endwin);
	} else {
		FindPoint(&x, &y, begwin, radius);
		x1int = round(x); y1int  =  round(y);
		arcwin = trunc(endwin - begwin);
		if (arcwin <= -180)
			arcwin = arcwin + 360;
		fprintf(druck_fp, "PU%i,%i;\n", x1int, y1int);
		fprintf(druck_fp, "PD;AA%i,%i,%i;PU;\n", xint, yint, arcwin);
	}
} /* DrawArc */



void DrawRoundVertex(double x1,  double y1,  double x2,  double y2,  double x3,  double y3,  double redlength)
	/* Rundet Ecken bei langen Kanten */
{
	int		mult;
	double	xmid, ymid, win, win1, win2, winrad, winred, winmid, radius, middlelength;
	double	radsin, redsin, midsin;

	mult = 1;
	FindAngle(x2, y2, x1, y1, &win1);
	FindAngle(x2, y2, x3, y3, &win2);
	x1 = x2; y1 = y2; x3 = x2; y3 = y2;
	FindPoint(&x1, &y1, win1, redlength);
	FindPoint(&x3, &y3, win2, redlength);
	win = win1 + 180;
	AdaptAngle(&win);
	if (win >= 180  &&  (win2 > win  ||  win2 < win1))
		win = win1 - win2;
	else if (win <= 180  &&  (win2 > win  &&  win2 < win1))
		win = win1 - win2;
	else {
		win = win2 - win1;
   		mult = -1;
	}
	AdaptAngle(&win);
	if (win > 178  &&  win < 182)
		DrawLine(x1, y1, x3, y3);
	else {
		winmid = 90;
		winrad = win / 2.0;
		winred = 90 - winrad;
		redsin = sin(winred * (double)PI / 180.0);
		radsin = sin(winrad * (double)PI / 180.0);
		midsin = sin(winmid * (double)PI / 180.0);
		radius = redlength * radsin / redsin;
		middlelength = redlength / redsin;
		win = win1 - mult * winrad;
		AdaptAngle(&win);
		xmid = x2; ymid = y2;
		FindPoint(&xmid, &ymid, win, middlelength);
		FindAngle(xmid, ymid, x1, y1, &win1);
		AdaptAngle(&win1);
		FindAngle(xmid, ymid, x3, y3, &win2);
		AdaptAngle(&win2);
		if (printer == FALSE)
			if (mult == 1)
				DrawArc(xmid, ymid, win1, win2, radius);
			else
				DrawArc(xmid, ymid, win2, win1, radius);
		else
			if (mult == 1)
				DrawArc(xmid, ymid, win1, win2, radius);
			else
				DrawArc(xmid, ymid, win1, win2, radius);
	}
} /* DrawRoundVertex */



void DrawLoop(double x,  double y, double win,  double redlength,  strg sym)
	/* Zeichne Schleife eines Zustandes,  wobei der Mittelpunkt des Zustandes
	und die Richtung (Winkel) angegeben wird. */
{
	int		textwidth;
	double	xarc, x1, x2, x3, x4, yarc, y1, y2, y3, y4;
	double	len, radius, rwin, lwin;
	double	xsym, ysym, quadrant, quadsqr;
	int		intwin;

	xarc = x1 = x3 = x;
	yarc = y1 = y3 = y;
	len = 7.0 / 4.0 * redlength;
	radius = round(redlength / 3.0) + (double)LOOPRADOFFS;
	FindPoint(&xarc, &yarc, win, len);
	xarc = round(xarc); yarc = round(yarc);
	x2 = xarc; y2 = yarc;
	x4 = xarc; y4 = yarc;
	rwin = win - 90;
	if (rwin < 0)
		rwin += 360;
	FindPoint(&x2, &y2, rwin, radius);
	lwin = win + 90;
	if (lwin > 360)
		lwin -= 360;
	FindPoint(&x4, &y4, lwin, radius);
	DrawArc(xarc, yarc, rwin, lwin, radius);
	rwin = win - 12;
	if (rwin < 0)
		rwin += 360;
	FindPoint(&x1, &y1, rwin, redlength);
	DrawArrow(x2, y2, x1, y1, radius);
	lwin = win + 12;
	if (lwin > 360)
		lwin -= 360;
	FindPoint(&x3, &y3, lwin, redlength);
	DrawLine(x3, y3, x4, y4);
	xsym = xarc; ysym = yarc;
	FindPoint(&xsym, &ysym, win, radius);
	quadrant = round(win / 90.0);
	FindTextLength(sym, &textwidth);
	intwin = trunc(win);
	if (intwin == 0  ||  intwin == 360) {
		xsym += FTBGOFFS;
		ysym -= (double)TEXTHEIGHT / 2.0;  
	} else if (intwin == 180) {
		xsym -= textwidth + FTBGOFFS;
		ysym -= (double)TEXTHEIGHT / 2.0;  
	} else if (intwin == 90) {
		xsym -= (double)textwidth / 2.0;
		ysym += FTBGOFFS;
	} else if (intwin == 270) {
		xsym -= (double)textwidth / 2.0;
		ysym -= TEXTHEIGHT + FTBGOFFS;
	} else {
		if (90 < intwin  &&  intwin < 270)
			xsym -= textwidth + FTBGOFFS;
		else
			xsym += FTBGOFFS;
		if (0 < intwin  &&  intwin < 180) {
			ysym += FTBGOFFS;
		} else {
			ysym -= TEXTHEIGHT + FTBGOFFS;
		}
	}
	DrawSymbol(xsym, ysym, sym);
} /* DrawLoop */



static void DrawInvArc(double x, double y, double begwin, double endwin, double radius)
{
	DrawArc(x, y, endwin, begwin, radius);
} /* DrawInvArc */



static void DrawInvArcArrow(double x, double y, double begwin, double endwin, double radius, double arrowlen)
{
	double xarbeg, yarbeg, xarend, yarend, arwin;

	xarend = x;
	yarend = y;
	FindPoint(&xarend, &yarend, endwin, radius);
	arwin = 2 * arctan(arrowlen / 2.0 * sqrt(1.0 / (16.0 * radius * radius - arrowlen * arrowlen)));
	arwin += endwin;
	AdaptAngle(&arwin);
	xarbeg = x;
	yarbeg = y;
	FindPoint(&xarbeg, &yarbeg, arwin+1, radius - 0.07);
	DrawArrow(xarbeg, yarbeg, xarend, yarend, arrowlen);
	DrawInvArc(x, y, begwin, endwin, radius);
} /* DrawInvArcArrow */



void DrawCurvedArrow(double x,  double y,  double x1,  double y1,  double radius, 
					double redlength,  double arrowlength,  strg symbol)
	/* Zeichne gekruemmte Linie zwischen zwei Punkten (x, y) und (x1, y1).
	radius>0  =>  Uhrzeigersinn
	radius < 0  =>  entgegengesetzter Uhrzeigersinn */
{
	double len, rad, mlen;
	double win, win1, helpwin, begwin, endwin, beg1win, end1win;
	double xarc, yarc, x1arc, y1arc, xbeg, ybeg, xend, yend, x1beg, y1beg, x1end, y1end;

	FindLength(x, y, x1, y1, &len);
	rad = abs((int)radius);						/* Abstand P - P1 < 2 * Radius   => Zeichne Kreis */
	if (len < 2 * rad)
		rad = round(len / 2.0);
	mlen = sqrt(rad * rad - redlength * redlength);
	FindAngle(x, y, x1, y1, &win);                       /* Winkel win : P -> P1 */
	win = round(win);
	FindAngle(x1, y1, x, y, &win1);                      /* Winkel win1: P1 -> P */
	win1 = round(win1);
	helpwin = win + 90;
	AdaptAngle(&helpwin);
	xarc = x; yarc = y;
	FindPoint(&xarc, &yarc, win, mlen);
	xbeg = x; ybeg = y;
	FindPoint(&xbeg, &ybeg, helpwin, redlength);
	FindAngle(xarc, yarc, xbeg, ybeg, &begwin);
	xend = xarc; yend = yarc;
	FindPoint(&xend, &yend, helpwin, rad);
	/* FindAngle(xarc, yarc, x, y, &endwin); */
	endwin = helpwin;
	x1arc = x1; y1arc = y1;
	FindPoint(&x1arc, &y1arc, win1, mlen);
	x1end = x1; y1end = y1;
	FindPoint(&x1end, &y1end, helpwin, redlength);
	FindAngle(x1arc, y1arc, x1end, y1end, &end1win);
	x1beg = x1arc; y1beg = y1arc;
	FindPoint(&x1beg, &y1beg, helpwin, rad);
	/* FindAngle(x1arc, y1arc, x1, y1, &beg1win); */
	beg1win = helpwin;
	DrawInvArc(xarc, yarc, begwin, endwin, rad);
	DrawLine(xend, yend, x1beg, y1beg);
	FindTextBegPoint(&xend, &yend, x1beg, y1beg, symbol);
	DrawSymbol(xend, yend, symbol);
	DrawInvArcArrow(x1arc, y1arc, beg1win, end1win, rad, arrowlength);
} /* DrawCurvedArrow */
