/* 
   Bordeaux : 1995
   */

#include "ext.h"
#include "extbis.h"

#include "win_una2.h"
#include "fa_utils.h"
#include "adjust.h"
#include "codetest.h"

#include <setjmp.h>


/* test if the cardinal of a language is different of 1 */
static boole cardinal_not_one(dfa X_dfa);

/* generic function to test if a code of DFA X_dfa  is 
   a prefix or suffix code :
   prefix : TRUE if test on prefix code 
            FALSE if test on suffix code */
static boole generic_xxxfix_code(dfa X_dfa, boole prefix);

/* generic function to test if a code of NFA X_nfa and DFA X_dfa is 
   a p_infix or s_infix code :
   p_infix : TRUE if test on p_infix code 
             FALSE if test on s_infix code */
static boole generic_infix_code(nfa X_nfa, dfa X_dfa, boole p_infix);

/* generic function for the bifix and infix code properties :
   if X_nfa == NULL : test the bifix code propertie
   else test the infix propertie */
static boole generic_bifix_infix_code(nfa X_nfa, dfa X_dfa, 
				      boolx prop1, boolx prop2);

/* generic function of tests on the semaphore codes :
   right  : TRUE if the test is about right_semaphore code
            FALSE if the test is about left_semaphore code
   xxxfix : required propretie for the semaphore codes (prefix or suffix) */
static boole generic_semaphore_code(nfa X_nfa,dfa X_dfa,boole right,boolx xxxfix);
				    
/* recursive deep scan in a DFA :
   state = state where we are in the scan of the automaton 
   length_path = length of the path followed to reach this state */
static recursive_deep(posint state, int length_path);

/* local variables for the functions recursive_deep and codetest_uniform_code */
posint  global_init;  /* initial state of the automaton scaned by recursive_deep */
posint  global_sno;   /* number of letters of the alphabet of this automaton */
posint  global_well;  /* well state of this automaton */
ddelta  global_delta; /* transition function of this automaton */
mrkfin  global_final; /* array of final states of this automaton */
jmp_buf buff;         /* used to stop the recursive scan of this automaton */
int     first_word_length; /* length of the first word encountered in this scan */
int    *global_length_path;/* length of the first path followed to reach a state */
boole  *visited;            /* array to mark visited states */

/* definitions for the functions recursive_deep and codetest_uniform_code */
#define NOT_UNIFORM 1
#define UNIFORM 2 
#define FIRST_WORD_LENGTH_INIT -1


/* ---------------------------- Global functions ------------------------------- */

boole codetest_code(dfa X_dfa)
/* The algorithm used is the Sardinas and Patterson Algorithm :
   (X is the code of DFA X_dfa)
   X_1 = X(-1)X - {epsilon}
   X_(i+1) = X(-1)X_i Union X_i(-1)X 
   the set of the X_i is finite
   X is a code if none of the X_i contains the empty word (epsilon) */
{ 
  posint sno; 

  sno = X_dfa->sno;

  if ( fa_utils_test_eps_dfa(X_dfa) == TRUE ) 
    /* Every language with epsilon isn't a code */
    return FALSE;
  else
    {
      dfa   tmp_dfa;
      dfa   eps_dfa;
      dfa   X_i_dfa;
      boole quit_true  = FALSE;
      boole quit_false = FALSE;
      
      eps_dfa = fa_utils_eps_dfa(sno);                           /* { eps } */
      tmp_dfa = adjust_leftquot_dfa(X_dfa, X_dfa);               /* X(-1).X */ 
      X_i_dfa = dfamdfa(adjust_minus_dfa(tmp_dfa, eps_dfa), TRUE);
                                                       /* X(-1).X - { eps } */
      freedfa(tmp_dfa);
      
      if ( fa_utils_test_empty_dfa(X_i_dfa) == TRUE )
	quit_true = TRUE;
      else
	{
	  dfa union_dfa;

	  union_dfa = fa_utils_empty_dfa(sno);

	  while( quit_true == FALSE && quit_false == FALSE )
	    {
	      dfa tmp1_dfa;
	      dfa tmp2_dfa;
	      dfa tmp3_dfa;
	      dfa tmp4_dfa;

	      tmp1_dfa  = union_dfa;
	      union_dfa = adjust_union_dfa(union_dfa, X_i_dfa); 
                                                              /* union U X_i */
	      tmp2_dfa  = adjust_leftquot_dfa(X_i_dfa, X_dfa);/* (X_i)(-1).X */
	      tmp3_dfa  = adjust_leftquot_dfa(X_dfa, X_i_dfa);/* X(-1).X_i   */
	      tmp4_dfa  = X_i_dfa;
	      X_i_dfa   = adjust_union_dfa(tmp2_dfa, tmp3_dfa);

	      fa_utils_free_temporary_fa(FALSE, 4, tmp1_dfa, tmp2_dfa, 
					 tmp3_dfa, tmp4_dfa);
	      
	      if( fa_utils_test_if_eps_in_dfa(X_i_dfa) == TRUE )
		quit_false = TRUE;  /* eps in X_i */
	      else
		if( adjust_inclusion_dfa(X_i_dfa, union_dfa) == TRUE )
		  quit_true = TRUE; /* X_i in union */	      
	    }
	  freedfa(union_dfa);
	}
      
      fa_utils_free_temporary_fa(FALSE, 2, eps_dfa, X_i_dfa);
      
      if ( quit_false == TRUE )
	return FALSE;
      else
	return TRUE;
    }
}



boole codetest_stack_code(dfa X_dfa, nfa X_nfa)
/* Algorithm :
   (X is the code of DFA X_dfa and NFA X_nfa)
   (X# is the stack words language (win_una2_stack_language))
   X is a #-code (stack_code) iff :
   for all u in Root(X#) : Card( u(-1)X Inter X# ) = 1
   thus :
   for any right residual u_i ( 0 <= i <= qno) of X, we verify that 
   Card(u_i(-1)X Inter X#) = 1 and if not for a given i, we verify 
   that Card(u_i(-1).X Inter Root(X#) = 0
   */
{
  if ( fa_utils_test_empty_dfa(X_dfa) == TRUE )
    return TRUE; 

  if ( fa_utils_test_if_eps_in_dfa(X_dfa) ) 
    /* Every code with epsilon isn't a stack-code */
    return FALSE;
  else
    {
      dfa    X_stack_dfa;      /* language of stack words of X */
      nfa    X_stack_nfa;      /* language of stack words of X */
      dfa    root_X_stack_dfa; /* root of the code of stack words of X */
      int    qno;     /* number of states of X_dfa */
      posint well;    /* well state of X_dfa */
      int    i;
      boole  result = TRUE;
      
      well    = fa_utils_well(X_dfa);
      qno     = X_dfa->qno;
      
      X_stack_dfa      = dfamdfa(win_una2_stack(X_dfa), TRUE);
      X_stack_nfa      = dfa2nfa(X_stack_dfa);
      root_X_stack_dfa = dfamdfa(win_una2_root(X_stack_dfa, X_stack_nfa), TRUE);
      
      for ( i = 0 ; i <= qno ; i++ )
	{
	  if ( i != well )
	    {
	      dfa   suffixdfa;
	      dfa   tmp1_dfa;
	      boole test;
	      
	      suffixdfa = fa_utils_suffix_i_dfa(X_dfa, i);
                                                     /* suffixdfa = u_i(-1)X */
	      tmp1_dfa  = adjust_insec_dfa(suffixdfa, X_stack_dfa); 
                                                           /* suffix inter X */
       	      test = cardinal_not_one(tmp1_dfa);
	      
	      fa_utils_free_temporary_fa(FALSE, 2, suffixdfa, tmp1_dfa);
	      
	      if ( test == TRUE )
		{
		  dfa   prefixdfa;
		  boole exit;
		  
		  prefixdfa = fa_utils_prefix_i_dfa(X_dfa, i); 
		  /* prefixdfa = class of the words u_i having the same u_i(-1)X */
		  exit = adjust_disjunction_dfa(prefixdfa, root_X_stack_dfa);
       		                           
		  freedfa(prefixdfa);
		  
		  if ( exit == FALSE )   /* prefix inter root(X#) != { } */
		    {		  
		      result = FALSE;
		      break;
		    }
		}	 
	    }
	}
      
      fa_utils_free_temporary_fa(FALSE, 2, X_stack_dfa, root_X_stack_dfa);

      return result;
    }
}


boole codetest_prefix_code(dfa X_dfa)
/* Algorithm :
   (X is the code of DFA X_dfa)
   X is a prefix code iff :
   X(-1)X = {epsilon}
   */
{
  return generic_xxxfix_code(X_dfa, TRUE);
}



boole codetest_suffix_code(dfa X_dfa)
/* Algorithm :
   (X is the code of DFA X_dfa)
   X is a suffix code iff :
   XX(-1) = {epsilon}
   */
{
  return generic_xxxfix_code(X_dfa, FALSE);
}



boole codetest_bifix_code(dfa X_dfa, boolx prefix, boolx suffix)
/* Algorithm :
   (X is the code of DFA X_dfa)
   X is a bifix code iff X is a prefix code and X is a suffix code 
   */

{  
  return generic_bifix_infix_code(NULL, X_dfa, prefix, suffix);
}



boole codetest_p_infix_code(nfa X_nfa, dfa X_dfa)
/* Algorithm :
   (X is the code of NFA X_nfa and DFA X_dfa)
   X is p_infix code on the alphabet A iff :
   ((A*).X.(A+)) Inter X = empty_set
   */
{
  return generic_infix_code(X_nfa, X_dfa, TRUE);
}



boole codetest_s_infix_code(nfa X_nfa, dfa X_dfa)
/* Algorithm :
   (X is the code of NFA X_nfa and DFA X_dfa)
   X is s_infix code on the alphabet A iff :
   ((A+).X.(A*)) Inter X = empty_set
   */
{
  return generic_infix_code(X_nfa, X_dfa, FALSE);
}



boole codetest_infix_code(nfa X_nfa, dfa X_dfa, boolx p_infix, boolx s_infix)
/* Algorithm :
   (X is the code of NFA X_nfa and DFA X_dfa)
   X is infix code on the alphabet A iff :
   X is a p_infix code and a s_infix code on A
   */
{
  return generic_bifix_infix_code(X_nfa, X_dfa, p_infix, s_infix);
}




boole codetest_outfix_code(dfa X_dfa)
/* Algorithm : (A is the alphabet)
   (X is the code of DFA X_dfa)
   for each state i of the automaton, we verify that :
   prefix_i_dfa(X, i).(A+).suffix_i_dfa(X, i) Inter X = empty_set
   */
{
  if ( fa_utils_test_empty_dfa(X_dfa) == TRUE )
    return TRUE; 
  else
    {
      dfa    alphabet_dfa;      /* A if A is the alphabet of X */
      dfa    plus_alphabet_dfa; /* A+ */
      posint qno;
      posint sno;
      posint well;
      int    i;
      boole  result = TRUE;
      
      sno     = X_dfa->sno;
      qno     = X_dfa->qno;
      well    = fa_utils_well(X_dfa); /* we assume that X_dfa is minimal */
      
      alphabet_dfa      = fa_utils_alphabet_dfa(sno);
      plus_alphabet_dfa = adjust_plus_dfa(alphabet_dfa);
            
      for ( i = 0 ; i <= qno ; i++ )
	{
	  if ( i != well )
	    {
	      dfa   prefix_dfa;
	      dfa   suffix_dfa;
	      dfa   tmp1_dfa;
	      dfa   tmp2_dfa;
	      boole exit;
	      
	      prefix_dfa = fa_utils_prefix_i_dfa(X_dfa, i);
	      suffix_dfa = fa_utils_suffix_i_dfa(X_dfa, i);
	      tmp1_dfa   = adjust_concat_dfa(prefix_dfa, plus_alphabet_dfa);
                                                       /* prefix.(A+) */
	      tmp2_dfa   = adjust_concat_dfa(tmp1_dfa, suffix_dfa);
                                                       /* prefix.(A+).suffix */
	      
	      exit = adjust_disjunction_dfa(tmp2_dfa, X_dfa);
	      
	      fa_utils_free_temporary_fa(FALSE, 4,prefix_dfa, 
					 suffix_dfa, tmp1_dfa, tmp2_dfa);
	      
	      if ( exit == FALSE )
		/* (prefix.(A+).suffix) inter X =/= { } */
		{                 
		  result = FALSE;
		  break;
		}
	    }
	}
      
      fa_utils_free_temporary_fa(FALSE, 2, alphabet_dfa, plus_alphabet_dfa);
      
      return result; 
    }
}


boole codetest_hypercode(dfa X_dfa, nfa X_nfa)
/* Algorithm :
   (X is the code of NFA X_nfa and of DFA X_dfa)
   X is a hypercode on the alphabet A iff :
   shuffle(X, (A+)) Inter X = empty_set
   */
{
  if ( fa_utils_test_empty_dfa(X_dfa) == TRUE )
    return TRUE; 
  {
    nfa   alphabet_nfa;         /* A */ 
    nfa   plus_alphabet_nfa;    /* A+ */
    nfa   tmp1_nfa;
    dfa   tmp_dfa;
    boole result;
    
    alphabet_nfa      = fa_utils_alphabet_nfa(X_dfa->sno); 
    plus_alphabet_nfa = starnfa(alphabet_nfa, TRUE);       
        
    tmp1_nfa = shuffle(X_nfa, plus_alphabet_nfa);     /* Shuffle(X,A+) */
    tmp_dfa  = nfa2dfa(tmp1_nfa);

    result = adjust_disjunction_dfa(tmp_dfa, X_dfa);/* (A*).Shuffle(X,A+)={ } */
    
    freedfa(tmp_dfa);
    fa_utils_free_temporary_fa(TRUE,3,tmp1_nfa,alphabet_nfa,plus_alphabet_nfa);
			       
    
    return result;
  }
}

  
boole codetest_uniform_code(dfa X_dfa)
/* Algorithm : 
   we verify that all the words recognized by X_dfa have the same length :
   we make a recursif deep scan of this automaton :
   if two words don't have the same length or if two paths of different
   length reach the same state : return FALSE */
{
  posint qno;
  int    i;
  int    result;

  qno          = X_dfa->qno;
  global_delta = X_dfa->delta;
  global_final = X_dfa->final;
  global_sno   = X_dfa->sno;
  global_well  = fa_utils_well(X_dfa);
  global_init  = X_dfa->init;

  first_word_length = FIRST_WORD_LENGTH_INIT;

  visited = (boole *) malloc((qno + 1) * sizeof(boole));
  if( visited == NULL )
    free_resources(1);

  global_length_path = (int *) malloc((qno + 1) * sizeof(int));
  if( global_length_path == NULL )
    {
      free(visited);
      free_resources(1);
    }
    
  for( i = 0 ; i <= qno ; i++ )
    {
      visited[i] = FALSE;
      global_length_path[i] = qno + 1;
    }

  result = setjmp(buff);
  if( result == 0 )
    recursive_deep(global_init, 0);
  else
    {
      free(visited);
      free(global_length_path);
      if( result == UNIFORM )
	return TRUE;
      else
	return FALSE;
    }
}


boole codetest_full_uniform_code(dfa X_dfa)
/* Algorithm : (A is the alphabet)
   we verify that all the words recognized by X_dfa have the same length 
   and that all the words of (A)* of this length are in X (language of DFA 
   X_dfa) : 
   for a state q ,all transition  must go to the same state. And if q is 
   the first final state encountered,all transition must go to the well */
{
  if( fa_utils_test_empty_dfa(X_dfa) == TRUE )
    return TRUE;
  else
    {
      posint well;
      posint q;
      posint sno;
      ddelta delta;
      mrkfin final;
      int    i;
      
      well  = fa_utils_well(X_dfa);
      delta = X_dfa->delta;
      final = X_dfa->final;
      sno   = X_dfa->sno;
      q = X_dfa->init;
      
      while( isfinal(final[q]) != TRUE )
	{
	  posint q0;
	  
	  q0 = delta[1][q];
	  if( q == q0 ) /*  there is a loop on q0 */
	    return FALSE;
	  else
	    {	  
	      for( i = 2 ; i <= sno ; i++ )
		if( delta[i][q] != q0 )
		  return FALSE;
	      q = q0;
	    } 
	}
      /* A^k is included in X */
      /* now, all the transitions must go to the well */
      
      for( i = 1 ; i <= sno ; i++ )
	if( delta[i][q] != well )
	  return FALSE;
      
      return TRUE;
    }
}


boole codetest_right_semaphore_code(nfa X_nfa, dfa X_dfa, boolx prefix)
/* Algorithm :
   (X is the code of NFA X_nfa and DFA X_dfa)
   X is a right semaphore code on the alphabet A iff:
   X is a prefix code and (A*).X C= X.(A*) (C= <==> in or equal)
   */
{
  return generic_semaphore_code(X_nfa, X_dfa, TRUE, prefix);
}


boole codetest_left_semaphore_code(nfa X_nfa, dfa X_dfa, boolx suffix)
/* Algorithm :
   (X is the code of NFA X_nfa and DFA X_dfa)
   X is a left semaphore code on the alphabet A iff:
   X is a suffix code and X.(A*) C= (A*).X (C= <==> in or equal)
   */
{
  return generic_semaphore_code(X_nfa, X_dfa, FALSE, suffix);
}


/* ---------------------------- Static functions ------------------------------- */


static boole cardinal_not_one(dfa X_dfa)
/* The cardinal of a language X (of DFA X_dfa) is different of one if :
   - there is more than one final state in the minimal DFA of X
   - there is no final state in the minimal DFA of X (empty language)
   - there is a cycle in the DFA : X has an infinity of words
   - there are at less 2 ways  to reach the same final state X has at 
     less 2 words */
{
  dfa    tmp_dfa;
  int    i, j;
  int    qno;
  int    sno;
  posint well;
  ddelta delta;
  mrkfin final;
  int    cpt = 0;    /* record the number of final states */
  boole  result = FALSE;
  boole *visited;

  if( X_dfa->minimal == TRUE )
    tmp_dfa = fa_utils_duplicate_dfa(X_dfa);
  else
    tmp_dfa = dfamdfa(X_dfa, FALSE);
  
  well    = fa_utils_well(tmp_dfa);
  delta   = tmp_dfa->delta;
  final   = tmp_dfa->final;
  sno     = tmp_dfa->sno;
  qno     = tmp_dfa->qno;

  visited = (boole *) malloc(qno * sizeof(boole));
  if( visited == NULL )
    {
      freedfa(tmp_dfa);
      free_resources(1);
    }

  for ( i = 0 ; i <= qno ; i++ )
    visited[i] = FALSE;


  for ( i = 0 ; i <= qno ; i++ )
    {
      if ( isfinal(final[i]) )
	cpt++;

      if ( cpt == 2 )      /* there are two final states : card(X) > 1 */
	break;
    }

  if ( cpt == 0 || cpt == 2 )          
    /* if cpt=0 there is no final state : card(X) = 0 */
    result = TRUE;
  else  /* there is 1 final state : 
           we search if there is a cycle in tmp_dfa or
           two ways to reach the  final  state */
    {
      for ( i = 0 ; i <= qno ; i++ )     
	{
	  if ( i != well )
	    {
	      boole one_meet = FALSE; /* to mark if we soon find a transition from
                                         i to an other state different of  well :
                                         thus there  are at less two ways to reach
                                         the final state */
	      visited[i] == TRUE;	      
	      for ( j = 1 ; j <= sno ; j++ )
		{
		  posint k;
		  
		  k = delta[j][i];   /* k = state reachable directly from i */
		  
		  if ( visited[k] == TRUE  )
		    /* there is a cycle in tmp_dfa : card(X) = infinite */
		    {
		      result = TRUE;
		      break;
		    }
		  else if ( k != well )       
		    if ( one_meet == FALSE )  
		      /* we find no transition to a state different of well */
		      one_meet = TRUE;        
		    else 
		      /* we find a transition to a second state different of well :
                         card(X) >= 2 */
		      {
			result = TRUE;
			break;
		      }
		}
	      
	      if ( result == TRUE )
		break;
	    }
	}
    }
  
  freedfa(tmp_dfa);
  free(visited);
  return result;     
}



static boole generic_xxxfix_code(dfa X_dfa, boole prefix)
{
  if( fa_utils_test_empty_dfa(X_dfa) == TRUE )
    return TRUE;
  else
    {
      dfa   tmp_dfa;
      dfa   eps_dfa;
      boole result;
      
      eps_dfa   = fa_utils_eps_dfa(X_dfa->sno);
      
      if ( prefix == TRUE )
       	tmp_dfa = dfamdfa(adjust_leftquot_dfa(X_dfa, X_dfa), TRUE);
      else
	tmp_dfa = dfamdfa(rightquot(X_dfa, X_dfa), TRUE);
      
      if( fa_utils_test_eps_dfa(tmp_dfa) == TRUE )
	result = TRUE;
      else
	result = FALSE;
      
      fa_utils_free_temporary_fa(FALSE, 2, eps_dfa, tmp_dfa);
      
      return result;
    }
}


static boole generic_infix_code(nfa X_nfa, dfa X_dfa, boole p_infix)
/* (X is the code of NFA X_nfa) */
{
  if ( fa_utils_test_empty_nfa(X_nfa) == TRUE )
    return TRUE; 
  else
    {
      nfa   alphabet_nfa;       /* A if A is the alphabet of X */
      nfa   plus_alphabet_nfa;  /* A+ */
      nfa   star_alphabet_nfa;  /* A* */
      nfa   tmp1_nfa;
      nfa   tmp2_nfa;
      dfa   tmp2_dfa;
      boole result;
      posint sno;
      
      sno = X_nfa->sno;
      
      alphabet_nfa      = fa_utils_alphabet_nfa(sno);
      star_alphabet_nfa = adjust_star_nfa(alphabet_nfa);
      plus_alphabet_nfa = adjust_plus_nfa(alphabet_nfa);
      
      if ( p_infix == TRUE )
	{
	  tmp1_nfa = concatfa(X_nfa, plus_alphabet_nfa);    /* X.(A+) */
	  tmp2_nfa = concatfa(star_alphabet_nfa, tmp1_nfa); /* (A*).X.(A+) */
	}
      else
	{
	  tmp1_nfa = concatfa(X_nfa, star_alphabet_nfa);    /* X.(A*) */
	  tmp2_nfa = concatfa(plus_alphabet_nfa, tmp1_nfa); /* (A+).X.(A*) */
	}
      tmp2_dfa = nfa2dfa(tmp2_nfa);

      result = adjust_disjunction_dfa(tmp2_dfa, X_dfa);
      
      freedfa(tmp2_dfa);
      fa_utils_free_temporary_fa(TRUE, 5, alphabet_nfa, plus_alphabet_nfa, 
				 star_alphabet_nfa, tmp1_nfa, tmp2_nfa);
      
      return result;
    }
}


boole generic_bifix_infix_code(nfa X_nfa, dfa X_dfa, boolx prop1, boolx prop2)
{
  boole tmp_prop1, tmp_prop2;

  if( prop1 == FALSE || prop2 == FALSE )
    return FALSE;

  if( prop1 == TRUE && prop2 == TRUE )
    return TRUE;  
  
  if( X_nfa == NULL ) /* test bifix code */
    {
      tmp_prop1 = (prop1 == UN_KNOWN ? codetest_prefix_code(X_dfa) : prop1);
      tmp_prop2 = (prop2 == UN_KNOWN ? codetest_suffix_code(X_dfa) : prop2);
    }
  else /* test infix code */
    {
      tmp_prop1 = (prop1 == UN_KNOWN ? codetest_p_infix_code(X_nfa,X_dfa) : prop1);
      tmp_prop2 = (prop2 == UN_KNOWN ? codetest_s_infix_code(X_nfa,X_dfa) : prop2);
    }    
  
  return ( tmp_prop1 && tmp_prop2 );
}


static boole generic_semaphore_code(nfa X_nfa, dfa X_dfa, 
				    boole right, boolx xxxfix)
/* (X is the code of NFA X_nfa and DFA X_dfa) */
{
  if ( fa_utils_test_empty_nfa(X_nfa) == TRUE )
    return TRUE; 

  if( xxxfix == FALSE )
    return FALSE;
  else
    {
      boole  result;
      boole  tmp_xxxfix; /* mark if the code is prefix or suffix */

      if ( xxxfix == UN_KNOWN )
	if ( right == TRUE )
	  tmp_xxxfix = codetest_prefix_code(X_dfa);
	else
	  tmp_xxxfix = codetest_suffix_code(X_dfa);
      else
	tmp_xxxfix = xxxfix;

      if ( tmp_xxxfix == FALSE )
	result = FALSE;
      else
	{
	  nfa    alphabet_nfa;	    /* A if A is the alphabet of X */
	  nfa    star_alphabet_nfa; /* A* */
	  nfa    tmp1_nfa;
	  nfa    tmp2_nfa;
	  posint sno;
      
	  sno = X_nfa->sno;

	  alphabet_nfa = fa_utils_alphabet_nfa(sno);

	  star_alphabet_nfa = adjust_star_nfa(alphabet_nfa); /* B = A* */
      
	  tmp1_nfa = concatfa(X_nfa, star_alphabet_nfa); /* X.B */
	  tmp2_nfa = concatfa(star_alphabet_nfa, X_nfa); /* B.X */
      
	  if ( right == TRUE )  /* A*.X in X.A* */
	    result = adjust_inclusion_nfa(tmp2_nfa, tmp1_nfa); /* A*.X in X.A* */
	  else                 /* X.A* in A*.X */
	    result = adjust_inclusion_nfa(tmp1_nfa, tmp2_nfa);
	  
	  fa_utils_free_temporary_fa(TRUE, 4, alphabet_nfa, star_alphabet_nfa,
				     tmp1_nfa, tmp2_nfa);
	}
      
      return result;
    }
}



static recursive_deep(posint state, int length_path)
{
  posint j;

  visited[state] = TRUE;
  global_length_path[state] = length_path;

  for( j = 1 ; j <= global_sno ; j++ )
    {
      posint new_state;
      new_state = global_delta[j][state];

      if( new_state != global_well )  /* we avoid the well state */
	{
	  boole final = isfinal(global_final[new_state]);	  

	  if( final == TRUE )
	    {
	      if( first_word_length == FIRST_WORD_LENGTH_INIT )
		first_word_length = length_path + 1; 

	      if( (length_path + 1) != first_word_length ) 
		/* two words don't have the same length : return FALSE */
		longjmp(buff, NOT_UNIFORM);		
	    }


	  if( visited[new_state] == TRUE )
	    {
	      if( global_length_path[new_state] != (length_path + 1) )
		/* there are 2 paths of different length to reach this state : 
		   code not uniform */
		longjmp(buff, NOT_UNIFORM);
	    }
	  else
	    recursive_deep(new_state, length_path + 1);
	}
    }
  
  if( state == global_init ) /* all paths have been followed */
    longjmp(buff, UNIFORM);  
}




