/* comp.c 3.0 (KIEL) 11.94 */ 
   
#include "ext.h"

/* _ID : the clock starts with an ID-no, and only a call with the
 *       same ID-no can stop the clock
 */

#define DFA_ID	1
#define MDFA_ID	2
#define NFA_ID	3
#define ENFA_ID	4
#define MON_ID	5
#define REX_ID	6
#define SFX_ID	7
#define DCL_ID	8
#define MNFA_ID	9

/* functions included
 *
 *	void compdfa
 *	void compmdfa
 *	void compnfa
 *	void compmon
 *	void compdclass
 *	void comprex
 *      void compsfx
 */

/******************************************************************/
/******************************************************************/
void
compdfa()
/* compute the deterministic automaton */
{	if (curlan->bdfa) return;
	cpuclock(DFA_ID);
	curlan->saved = FALSE;
        compnfa();
	if (curlan->bnfa) {
		newtick("Transformation from NFA to DFA",0);
		curlan->ldfa = nfa2dfa(curlan->lnfa);
		curlan->bdfa = TRUE;
		cpuclock(DFA_ID);
		return;
	}
} /* compdfa */

/******************************************************************/
void
compmdfa()
/* computes the minimal dfa */
{
	if (curlan->bdfa) 
	     if (curlan->ldfa->minimal) return; 
	cpuclock(MDFA_ID);
	if (!curlan->bdfa) compdfa();
	if (curlan->inputis == DFAINP) 
               {
		curlan->odfa = curlan->ldfa;
		curlan->inputis = ODFAINP;
	       }
	 newtick("Transformation from DFA to minimal DFA",curlan->ldfa->qno+1);
	 curlan->ldfa = dfamdfa(curlan->ldfa,FALSE);
	 curlan->saved = FALSE;
 	 cpuclock(MDFA_ID);
} /* compmdfa */

/******************************************************************/
void
compnfa()
/* compute a nondeterministic finite automaton */
{
	if (curlan->bnfa) return;
	curlan->saved = FALSE;
	if (curlan->bdfa) 
               {
		curlan->lnfa = dfa2nfa(curlan->ldfa);
		curlan->bnfa = TRUE;
		return;
	       }
	else if (curlan->inputis == OENFAINP)
               {
		cpuclock(NFA_ID);
	  	newtick("Transformation from epsilon-NFA to NFA",
		((curlan->sizeal * 2 + 1) * (curlan->oenfa->qno + 1)));
		curlan->lnfa = enfa2nfa(curlan->oenfa);
		curlan->bnfa = TRUE;
		cpuclock(NFA_ID);
		return;
	       }
	else if ((curlan->brex) && (!curlan->lrex->grex)) 
               {
		cpuclock(NFA_ID);
		newtick("Transf. from Regular Expression to NFA",
			curlan->lrex->erexl*3);
		curlan->lnfa = rex2nfa(curlan->lrex);
		curlan->bnfa = TRUE;
		cpuclock(NFA_ID);
		return;
	       }
	else if (curlan->brex) 
               {
		cpuclock(NFA_ID);
	  	newtick("Transf. from Gen. Regular Expression to NFA",
			curlan->lrex->erexl);
		curlan->lnfa = genrex2nfa(curlan->lrex);
		curlan->bnfa = TRUE;
		cpuclock(NFA_ID);
		return;
	       }
} /* compnfa */

/******************************************************************/
void
compmnfa()
/* compute a nondeterministic minimal finite automaton */
{
 nfa help;
        if(curlan->bnfa)
             if(curlan->lnfa->minimal)return;
	if (curlan->inputis == NFAINP) 
           { 
            curlan->inputis = ONFAINP; 
            curlan->onfa = curlan->lnfa;
           }
        cpuclock(MNFA_ID);
        compmdfa();
        curlan->saved = FALSE;
        newtick("Transformation from NFA to minimal NFA",0);
        if(!curlan->bnfa)
          help=dfa2nfa(curlan->ldfa);
        else
          help=curlan->lnfa;
        curlan->lnfa = nfa2mnfa(help,curlan->ldfa);
        curlan->lnfa->minimal=TRUE;
        curlan->bnfa=TRUE;
        cpuclock(MNFA_ID);
        return;
} /* compmnfa */
/******************************************************************/

void
compmon()
/* compute the monoid structure */
{ 
if(!curlan->bmon)
     {
      cpuclock(MON_ID);
      compmdfa();
      newtick("Transformation from minimal DFA to Monoid",0);
      curlan->lmon=dfa2mon(curlan->ldfa);
      curlan->bmon=TRUE;
      curlan->saved = FALSE;
      cpuclock(MON_ID);
     }
} /* compmon */

/******************************************************************/
void compdclass()
{
        if(curlan->bmon)
	  if (curlan->lmon->dclassiscomputed) return; 
 	cpuclock(DCL_ID);
	compmon();
        newtick("Computation of the D-class structure",curlan->lmon->mno);
	mon2dcl(curlan->lmon);
	curlan->saved = FALSE;
	cpuclock(DCL_ID);
} /* compdclass */

/******************************************************************/
void
comprex()
/* compute a regular expression from curlan */
{ boole nocomp=curlan->brex;
  nfa   mnfa;
	if ((curlan->brex) && (curlan->lrex->grex)) 
               { /* copy original input */
		curlan->orex = curlan->lrex;
		curlan->inputis = OREXINP;
		nocomp = FALSE;
	       }
	if(!nocomp) {
	        cpuclock(REX_ID);
		curlan->saved = FALSE;
                /* search for minimal automaton */
                if(curlan->bdfa)
                  if(curlan->bnfa)
                    {
                     compmdfa();
                     if(curlan->ldfa->qno <= curlan->lnfa->qno)
                       mnfa=dfa2nfa(curlan->ldfa);
                     else
                       mnfa=curlan->lnfa;
                    }
                  else 
                    {
                     compmdfa();
                     mnfa=dfa2nfa(curlan->ldfa);
                    }
                else
		  {
                   compnfa();
                   mnfa=curlan->lnfa;
                  }
   		newtick("Transformation from NFA to Regular Expression",
			mnfa->qno + 3);
		curlan->lrex = nfa2rex(mnfa);
	        curlan->brex = TRUE;
	        cpuclock(REX_ID);
	}
} /* comprex */

/******************************************************************/
void
compsfx()
{ 
 if(curlan->stariscomp)return;
 cpuclock(SFX_ID); 
 compmon();
 if(curlan->starfree == UN_KNOWN) 
   {curlan->starfree=sftest(curlan->lmon);   /* test if starfree */
    if(curlan->starfree == TRUE) curlan->folu=TRUE;
    curlan->saved = FALSE;
   }
 if(curlan->starfree) 
   {newtick("computation of a starfree expression",curlan->lmon->mno);
    curlan->sexp =mon2sfx(curlan->ldfa,curlan->lmon);
    curlan->stariscomp = TRUE;
    curlan->saved = FALSE;
   }
 cpuclock(SFX_ID);
} /* compsfx */
