#ifndef DEF_CODE_H
#define DEF_CODE_H

/* 
   Bordeaux : 1995
   */

#define CODE \
"Code:
  A language L is a code if every word of L* has at most one 
  factorsation in words of L."

/* Definitions of properties on the codes */

#define PREFIX \
"Prefix code : 
  L is a prefix code on an alphabet A if LQ(L,L) = { eps },
  where LQ is the left quotient and eps the empty-word, i.e :
  no word of L is a strict prefix of an other word of L.
  If L is a prefix code and a suffix code, L is a bifix code."

#define SUFFIX \
"Suffix code :
  L is a suffix code on an alphabet A if RQ(L,L) = { eps },
  where RQ is the rightt quotient and eps the empty-word, i.e :
  no word of L is a strict suffix of an other word of L.
  If L is a prefix code and a suffix code, L is a bifix code."

#define BIFIX \
"Bifix code:
  L is a bifix code if it is both a prefix code and a suffix code."

#define P_INFIX \
"P-infix code : 
  L is a p-infix code on an alphabet A if, for all x,u,y in A*,
  xuy in L and u in L together imply y = empty-word. 
  If L is a p-infix code and a s-infix code, L is an infix code. 
 --------------- 
 See \"Outfix and infix codes and related classes of languages\", 
     Ito, Jurgensen, Shyr, Thierrin, 
     Journal of Computer and System Sciences, 1991, p 484-508"

#define S_INFIX \
"S-infix code : 
  L is a s-infix code on an alphabet A if, for all x,u,y in A*, 
  xuy in L and u in L together imply x = empty-word. 
  If L is a p-infix code and a s-infix code, L is an infix code. 
 --------------- 
 See \"Outfix and infix codes and related classes of languages\", 
     Ito, Jurgensen, Shyr, Thierrin, 
     Journal of Computer and System Sciences, 1991, p 484-508"

#define INFIX \
"Infix code:
  L is an infix code if it is both a p-infix code and a s-infix code.
  L is an infix code on an alphabet A if ,for all x,u,y in A*,
  xuy in L and u in L together imply xy = empty-word. 
 --------------- 
 See \"Outfix and infix codes and related classes of languages\", 
     Ito, Jurgensen, Shyr, Thierrin, 
     Journal of Computer and System Sciences, 1991, p 484-508"

#define OUTFIX \
"Outfix code: 
  L is an outfix code on an alphabet A if for all x, y, u in A*, 
  xy in L and xuy in L together imply u = empty-word. 
 --------------- 
 See \"Outfix and infix codes and related classes of languages\", 
     Ito, Jurgensen, Shyr, Thierrin, 
     Journal of Computer and System Sciences, 1991, p 484-508"

#define HYPERCODE \
"Hypercode (on an alphabet A): 
  L is an hypercode on A if ( Shuffle(L,A+) inter L ) = empty-set.
 --------------- 
 See \"Outfix and infix codes and related classes of languages\", 
     Ito, Jurgensen, Shyr, Thierrin, 
     Journal of Computer and System Sciences, 1991, p 484-508"

#define R_SEMAPHORE \
"Right semaphore code:
  L is a right semaphore code on an alphabet A if L is a prefix 
  code on A satisfying (A*).L included in L.(A*).
 --------------- 
 See \"Outfix and infix codes and related classes of languages\", 
     Ito, Jurgensen, Shyr, Thierrin, 
     Journal of Computer and System Sciences, 1991, p 484-508"

#define L_SEMAPHORE \
"Left semaphore code:
  L is a left semaphore code on an alphabet A if L is a suffix
  code on A satisfying L.(A*) included in (A*).L.
 ---------------
 See \"Outfix and infix codes and related classes of languages\", 
     Ito, Jurgensen, Shyr, Thierrin, 
     Journal of Computer and System Sciences, 1991, p 484-508"

#define UNIFORM \
"Uniform code:
  L is an uniform code on an alphabet A if for all x,y in L, 
  |x| = |y|.
 ---------------
 See \"Outfix and infix codes and related classes of languages\", 
     Ito, Jurgensen, Shyr, Thierrin, 
     Journal of Computer and System Sciences, 1991, p 484-508" 

#define FULL_UNIFORM \
"Full uniform code:
  L is a full uniform code on an alphabet A if  L = A^k for k 
  positve, where k = |u| for a word u in L.
  In other words if for all x,y in L, |x| = |y| and there is no 
  word z in A* such that there is a word  w in L where |w| = |z| 
  and z not in L.

  The full uniform propertie is equivalent to the strict right
  semaphore propertie (L is prefix and (A+).L in L.(A+)) and
  to the strict left semaphore propertie (L is suffix and 
  L.(A+) in (A+).L).
 ---------------
 See \"Outfix and infix codes and related classes of languages\", 
     Ito, Jurgensen, Shyr, Thierrin, 
     Journal of Computer and System Sciences, 1991, p 484-508"

#define STACK_CODE \
"Stack code:
  For the notion of stack code :
  See \"On stack languages\", 
      B.Patrou I.Litovsky B.Le Saec, 
      Algebraic Theory and Applications to Formal Languages and 
      Codes, 1993, p 174-182"

char *definitions[] = 
{
  PREFIX,
  SUFFIX,
  BIFIX,
  P_INFIX,
  S_INFIX,
  INFIX,
  OUTFIX,
  HYPERCODE,
  R_SEMAPHORE,
  L_SEMAPHORE,
  UNIFORM,
  FULL_UNIFORM,
  STACK_CODE
};

#endif
