/*
   Bordeaux : 1995
   */

#include "ext.h"
#include "extbis.h"

#include "edcode.h"
#include "def_code.h"

#include <ctype.h>
#include <string.h>
#include <curses.h>
#include <stdlib.h>

#define START_COL 1  /* first column for displaying */
#define LEN_COLS  80 /* number of columns allowed */
#define ROWREF    3  /* first line of the part References of the language */
#define ROWPROP   7  /* first line of the part Properties of the language */
#define ENDROW    23 /* number of lines allowed */
#define PROPCOL   20

#define TITLE   "C O D E S  I N F O R M A T I O N S"
#define STITLE1 "L A N G U A G E  R E F E R E N C E S"
#define STITLE2 "C O D E S  P R O P E R T I E S"

#define QUIT1    "a..m : see the corresponding definition, q : quit"
#define QUIT2    "q : quit"
#define QUIT3    "h : left, l : right, q : quit"
#define RET      "r : return to the precedent window"
#define LEFT     'h'
#define RIGHT    'l'
#define QUIT     'q'
#define RETURN   'r'

#define SCROLL   "..."

static language global_lang; /* language to display */
static char *word_not_code;  /* word to eventually display */

static void make_subtitle(int row, char *txt);

/* display the value of a propertie of language */
static void prop(boole prop);

/* display the codes properties of the language */
static void display_code_properties();

/* display the two factorisations of the language */
static void display_two_factorisations();

/* display a (char *) with horozontal scrolling */
static void display_string_scroll(int row, int col, int dpct, char *string);

/* clear the properties window */
static void clear_prop_window();

/* display the definition of a propertie : number corresponds to the propertie */
static void display_def(int number);

/* in some systems, the function addnstr and mvaddnstr of the curses are not 
   available */
static void ed_mvaddnstr(int y, int x, char *str, int n);
static void ed_addnstr(char *str, int n);


/* the two factorisations of word_not_code and the word without separators */
static char *factorisation1; 
static char *factorisation2;
static char *word;

/* characters used to isolate the factors of factorisation1 and factorisation2 */
#define SEP1 '(' 
#define SEP2 ')'

/* build the words to be displayed in the case of displaying a double 
   factorisation */
static void filter_word_not_code(char *word_not_code);


#ifdef X_DISPLAY

#include <sys/types.h>
#include <unistd.h>

#include <X11/Intrinsic.h>	
#include <X11/StringDefs.h>
#include <X11/Shell.h>

#include <X11/Xaw/Cardinals.h>
#include <X11/Xaw/Box.h>	
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/List.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>


#define WIN_NAME "Codes Informations"
#define WIN_NAME_LENGTH 18

#define NB_CODE_PROP 13
#define PROP_LENGTH 27
#define ASK_SYMBOL_POS 24

String res[] = {
  "*code_label.label:           Code Properties",
  "*ref_text.scrollHorizontal:	whenNeeded",
  "*ref_text.height:            30",
  "*prop_view.allowVert:        True", 
  "*prop_list.verticalList:     True",
  "*prop_list.defaultColumns:   1",
  "*quit.label:            Quit",
  "*help_label.label:      Code Definitions",
  "*help_text.scrollVertical:	whenNeeded",
  "*help_text.scrollHorizontal:	whenNeeded",
  "*help_text.autoFill:	   on",
  "*help_text.width:       550",
  "*help_text.height:      100",
  "*non_code_label.label:       Two Factorisations",
  "*eps_in_dfa_label.label:     "EPS_IN_DFA,
  "*fact1_text.scrollVertical:	whenNeeded",
  "*fact1_text.scrollHorizontal:	whenNeeded",
  "*fact1_text.autoFill:	   on",
  "*fact1_text.height:      30",
  "*fact2_text.scrollVertical:	whenNeeded",
  "*fact2_text.scrollHorizontal:	whenNeeded",
  "*fact2_text.autoFill:	   on",
  "*fact2_text.height:      30",
  NULL
};	

String help_text_on_code_properties[NB_CODE_PROP] = {
  PREFIX,
  SUFFIX,
  BIFIX,
  P_INFIX,
  S_INFIX,
  INFIX,
  OUTFIX,
  HYPERCODE,
  UNIFORM,
  FULL_UNIFORM,
  R_SEMAPHORE,
  L_SEMAPHORE,
  STACK_CODE
};

String constant_code_properties_list[NB_CODE_PROP+1] = {
  "prefix                  ???   ",
  "suffix                  ???   ",
  "bifix                   ???   ",
  "p-infix                 ???   ",
  "s-infix                 ???   ",
  "infix                   ???   ",
  "outfix                  ???   ",
  "hypercode               ???   ",
  "uniform                 ???   ",
  "full uniform            ???   ",
  "right semaphore         ???   ",
  "left semaphore          ???   ",
  "stack-code              ???   ",
  NULL};

String code_properties_list[NB_CODE_PROP];

static int argc= 1;
static char *argv[1];

static void quit (Widget w, XtPointer client_data, XtPointer call_data);

static void help_code_properties (Widget w, XtPointer help_text, 
				  XtPointer call_data); 

static void set_code_property_to_Yes (int p);
/* Puts the code property to Yes to be displayed (not in the language) */

static void set_code_property_to_No (int p);
/* Puts the code property to No to be displayed (not in the language) */

static void set_code_property_to (int p, boolx cp);
/* Puts the code property to 'cp' to be displayed (not in the language) */
#endif

/* ---------------------------- Global functions ------------------------------- */


void edcode_display_codes_info(language L)
{
  global_lang = L;

  word_not_code = global_lang->word_not_code;

  clear(); 

  maketitle(TRUE, TITLE);
  
  /* first part of the window */
  make_subtitle(ROWREF,STITLE1);
  mvaddstr(ROWREF+1, START_COL, "Name          : ");
  addstr(global_lang->name);
  mvaddstr(ROWREF+2, START_COL, "Reference     : ");
  addstr(global_lang->ref);
  mvaddstr(ROWREF+3, START_COL, "Alphabet size : ");
  addstr(pi2a(global_lang->sizeal));

  /* second part of the window */
  make_subtitle(ROWPROP, STITLE2);
  
  if ( global_lang->code == TRUE )
    display_code_properties();
  else
    display_two_factorisations();

  drawmenu(menu);
}


#ifdef X_DISPLAY

void edcode_xdisplay_codes_info (language L) 
{
  pid_t pid;
  
  pid = fork();

  if ( pid == 0 ) 
    {
      int i;
      XtAppContext app_con;
      Widget top, display, name_label,
                          ref_text,
                          code_label,
                          prop_view, prop_box, prop_list, 
                          help_label,
                          help_text,
                          non_code_label,
                          eps_in_dfa_label,
                          fact1_text, 
                          fact2_text,
                          quit_box, quit_command;
      
      Arg name_args[1], ref_args[1], help_args[1],
      fact1_args[1], fact2_args[1];
      
      argv[0] = (char *) malloc ( sizeof(char) * (WIN_NAME_LENGTH + 1) );
      if( argv[0] == NULL )
	free_resources(1);
      strcpy (argv[0], WIN_NAME);
                  
      for ( i = 0 ; i < NB_CODE_PROP ; i++ )
	code_properties_list[i] = XtNewString (constant_code_properties_list[i]);
      
      top= XtAppInitialize (&app_con, "code_display", NULL, ZERO, &argc, 
			    argv, res, NULL, ZERO);
      display = XtCreateManagedWidget ("display", panedWidgetClass, top, 
				       NULL, ZERO);

      XtSetArg (name_args[0], XtNlabel, L->name);
      name_label = XtCreateManagedWidget("name_label", labelWidgetClass, display,
					 name_args, ONE);
      XtSetArg (ref_args[0], XtNstring, L->ref);
      ref_text = XtCreateManagedWidget("ref_text", asciiTextWidgetClass, display, 
				       ref_args, ONE);
    if ( L->code == TRUE ) 
      {
	set_code_property_to (0, L->prefix_code);
	set_code_property_to (1, L->suffix_code);
	set_code_property_to (2, L->bifix_code);
	set_code_property_to (3, L->p_infix_code);
	set_code_property_to (4, L->s_infix_code);
	set_code_property_to (5, L->infix_code);
	set_code_property_to (6, L->outfix_code);
	set_code_property_to (7, L->hypercode);
	set_code_property_to (8, L->uniform_code);
	set_code_property_to (9, L->full_uniform_code);
	set_code_property_to (10, L->right_semaphore_code);
	set_code_property_to (11, L->left_semaphore_code);
	set_code_property_to (12, L->stack_code);
	
	code_label = XtCreateManagedWidget("code_label", labelWidgetClass, display,
					   NULL, ZERO);
	prop_view = XtCreateManagedWidget("prop_view", viewportWidgetClass, 
					  display, NULL, ZERO);
	prop_box = XtCreateManagedWidget ("prop_box", boxWidgetClass, prop_view,
					  NULL, ZERO); 
	prop_list = XtCreateManagedWidget("prop_list", listWidgetClass, prop_box, 
					  NULL, ZERO);
	XawListChange (prop_list, code_properties_list, NB_CODE_PROP, 0, True);
	
	help_label = XtCreateManagedWidget("help_label", labelWidgetClass, display,
					   NULL, ZERO);
	
	XtSetArg (help_args[0], XtNstring, CODE);
	help_text = XtCreateManagedWidget ("help_text", asciiTextWidgetClass, 
					   display, help_args, ONE);
	XtAddCallback(prop_list, XtNcallback, help_code_properties, 
		      (XtPointer) help_text);
      }
    else
      {      
	if ( strcmp(L->word_not_code, EPS_IN_DFA) == 0 )
	  {
	    eps_in_dfa_label = XtCreateManagedWidget ("eps_in_dfa_label", 
						      labelWidgetClass, display, 
						      NULL, ZERO);
	  } 
	else 
	  {
	    non_code_label = XtCreateManagedWidget ("non_code_label", 
						    labelWidgetClass, display,
						    NULL, ZERO);
	    filter_word_not_code (L->word_not_code);
	    XtSetArg (fact1_args[0], XtNstring, factorisation1);
	    fact1_text = XtCreateManagedWidget ("fact1_text", asciiTextWidgetClass,
						display, fact1_args, ONE);
	    XtSetArg (fact2_args[0], XtNstring, factorisation2);
	    fact2_text = XtCreateManagedWidget ("fact2_text", asciiTextWidgetClass,
						display, fact2_args, ONE);
	  }
      }
      
      quit_box = XtCreateManagedWidget ("quit_box", boxWidgetClass, display,
					NULL, ZERO);
      
      quit_command = XtCreateManagedWidget("quit", commandWidgetClass, quit_box,
					   NULL, ZERO);
      
      XtAddCallback(quit_command, XtNcallback, quit, NULL);
      
      XtRealizeWidget(top);
      XtAppMainLoop(app_con);
      free(argv[0]);
    }
  else  
    if (pid == -1)
      {
	help(TRUE,"Unable to display the graphical window on codes. Press a key.");
	edcode_display_codes_info(L);
      }
}

#endif

/* ---------------------------- Static functions ------------------------------- */


static void make_subtitle(int row, char *txt)
{
  int length = strlen(txt) + 2;
  int rest   = LEN_COLS - length;
  int part1  = rest / 2;
  int part2  = rest - part1;
  int i;
  
  move(row, START_COL);
  for ( i = 1 ; i <= part1 - 1 ; i++ )
    addch('=');
  addch(' ');
  addstr(txt);
  addch(' ');
  for ( i = 1 ; i <= part2 ; i++ )
    addch('=');
  refresh();
}



static void prop(boole prop)
{
  switch(prop)
    {
    case TRUE:
      addstr("Yes");
      break;
    case FALSE:
      addstr("No");
      break;
    case UN_KNOWN:
      addstr("?");
      break;
    default:
      break;
    }
}



static void display_code_properties()
{
  char  choice = '/';
  boole quit = FALSE;
  
  while( quit == FALSE )
    {
      switch(choice)
	{
	case QUIT:
	  quit = TRUE;
	  break;
	case 'a':
	case 'b':
	case 'c':
	case 'd':
	case 'e':
	case 'f':
	case 'g':
	case 'h':
	case 'i':
	case 'j':
	case 'k':
	case 'l':
	case 'm':
	  display_def(choice - 'a');
	default:
	  mvaddstr(ROWPROP+2, START_COL, 
		   "This language is a code, it has the following properties :");  
	  mvaddstr(ROWPROP+3, START_COL+PROPCOL, 
		   "[a] prefix code --------------------> ");
	  prop(global_lang->prefix_code);
	  mvaddstr(ROWPROP+4, START_COL+PROPCOL, 
		   "[b] suffix code --------------------> ");
	  prop(global_lang->suffix_code);
	  mvaddstr(ROWPROP+5, START_COL+PROPCOL, 
		   "[c] bifix code ---------------------> ");
	  prop(global_lang->bifix_code);
	  mvaddstr(ROWPROP+6, START_COL+PROPCOL, 
		   "[d] p-infix code -------------------> ");
	  prop(global_lang->p_infix_code);
	  mvaddstr(ROWPROP+7, START_COL+PROPCOL, 
		   "[e] s-infix code -------------------> ");
	  prop(global_lang->s_infix_code);
	  mvaddstr(ROWPROP+8, START_COL+PROPCOL, 
		   "[f] infix code ---------------------> ");
	  prop(global_lang->infix_code);
	  mvaddstr(ROWPROP+9, START_COL+PROPCOL, 
		   "[g] outfix code --------------------> ");
	  prop(global_lang->outfix_code);
	  mvaddstr(ROWPROP+10, START_COL+PROPCOL, 
		   "[h] hypercode ----------------------> ");
	  prop(global_lang->hypercode); 
	  mvaddstr(ROWPROP+11, START_COL+PROPCOL, 
		   "[i] right semaphore code -----------> ");
	  prop(global_lang->right_semaphore_code);
	  mvaddstr(ROWPROP+12, START_COL+PROPCOL, 
		   "[j] left semaphore code ------------> ");
	  prop(global_lang->left_semaphore_code);
	  mvaddstr(ROWPROP+13, START_COL+PROPCOL, 
		   "[k] uniform code -------------------> ");
	  prop(global_lang->uniform_code);
	  mvaddstr(ROWPROP+14, START_COL+PROPCOL, 
		   "[l] full uniform code --------------> ");
	  prop(global_lang->full_uniform_code);
	  mvaddstr(ROWPROP+15, START_COL+PROPCOL, 
		   "[m] stack-code ---------------------> ");
	  prop(global_lang->stack_code);
	  mvaddstr(ENDROW, START_COL, QUIT1);
	  refresh();
	  choice = tolower(getch());	  
	  break;
	}
    }
}


static void display_two_factorisations()
{
  int i;
  char choice;

  if ( strcmp(word_not_code, EPS_IN_DFA) == 0)
    {
      mvaddstr(ROWPROP+2, START_COL, EPS_IN_DFA);
      mvaddstr(ENDROW, START_COL, QUIT2);      
      refresh();
      do
	choice = tolower(getch());
      while( choice != QUIT );
    }
  else
    {
      int   dpct = 0;
      boole scrolling;
      int len_fac1, len_fac2, len_word;
     
      filter_word_not_code(word_not_code);

      len_fac1 = strlen(factorisation1);
      len_fac2 = strlen(factorisation2);
      len_word = strlen(word);

      scrolling = (len_fac1 > LEN_COLS || len_fac2 > LEN_COLS || 
		   len_word > LEN_COLS);
		  		  
      mvaddstr(ROWPROP+2, START_COL, 
	       "This language is not a code, the following word has two factorisations :");
      mvaddstr(ROWPROP+6, START_COL, "Factorisation 1 : ");
      mvaddstr(ROWPROP+9, START_COL, "Factorisation 2 : ");
      mvaddstr(ENDROW, START_COL, QUIT3);      
      do
	{	  
	  display_string_scroll(ROWPROP+4, START_COL, dpct, word);
	  display_string_scroll(ROWPROP+7, START_COL, dpct, factorisation1);
	  display_string_scroll(ROWPROP+10, START_COL, dpct, factorisation2);
	  
	  switch( choice = tolower(getch()) )
	    {
	    case LEFT: 
	      if( scrolling == TRUE && dpct > 0 )
		dpct--;
	      break;
	    case RIGHT:
	      if ( scrolling == TRUE && (dpct < (len_fac1 - LEN_COLS) ||
					 dpct < (len_fac2 - LEN_COLS) || 
					 dpct < (len_word - LEN_COLS)) )		
		dpct++;
	      break;
	    case QUIT:
	    default :
	      break;
	    }
	}
      while( choice != QUIT );
    }
}


static void display_string_scroll(int row, int col, int dpct, char *string)
{
  int   len_str = strlen(string);
  int   len_scroll = strlen(SCROLL);
  int   len_txt;
  char *ptr;
  
  if( len_str < LEN_COLS )
    {
      mvaddstr(row, col, string);
      refresh();
      return;
    }

  if( dpct == 0 )
    {
      ed_mvaddnstr(row, col, string, LEN_COLS - len_scroll - 1);
      addstr(SCROLL);
      refresh();
      return;
    }
  
  mvaddstr(row, col, SCROLL);
  if( dpct >= len_str - LEN_COLS )
    {
      len_txt = LEN_COLS - len_scroll - 1;
      ptr = string + len_str - len_txt;
      ed_addnstr(ptr, len_txt);
    }
  else
    {
      ptr = string + dpct;
      len_txt = LEN_COLS - (2 * len_scroll) -1 ;
      ed_addnstr(ptr, len_txt);
      addstr(SCROLL);
    }
  refresh();
}
	
	
static void clear_prop_window()
{
  int i;
  
  for( i = ROWPROP+1 ; i <= ENDROW ; i++ )
    {
      move(i, START_COL);
      clrtoeol();
    }
  refresh();
}


static void display_def(int number)
{  
  char ret;

  clear_prop_window();
  mvaddstr(ROWPROP+2, START_COL, definitions[number]);
  mvaddstr(ENDROW, START_COL, RET);
  refresh();

  ret = tolower(getch());
  while( ret != RETURN )
    ret = getch();

  clear_prop_window();
}


static void ed_addnstr(char *str, int n)
{
  int i;

  for( i = 0 ; i < n ; i++ )
    addch(str[i]);
}
  


static void ed_mvaddnstr(int x, int y, char *str, int n)
{
  move(x, y);
  ed_addnstr(str, n);
}


#ifdef X_DISPLAY

static void help_code_properties (Widget w, XtPointer text_ptr, 
				  XtPointer call_data) 
{
  XawListReturnStruct *item = (XawListReturnStruct*)call_data;
  Widget help_text= (Widget) text_ptr;
  Arg args[1];
  
  XtSetArg(args[0], XtNstring, 
	   help_text_on_code_properties[item->list_index]);
  XtSetValues(help_text, args, ONE);
}

static void set_code_property_to_Yes (int p) 
{
  code_properties_list[p][ASK_SYMBOL_POS]=   'Y';
  code_properties_list[p][ASK_SYMBOL_POS+1]= 'e';
  code_properties_list[p][ASK_SYMBOL_POS+2]= 's';
}

static void set_code_property_to_No (int p) 
{
  code_properties_list[p][ASK_SYMBOL_POS]=   'N';
  code_properties_list[p][ASK_SYMBOL_POS+1]= 'o';
  code_properties_list[p][ASK_SYMBOL_POS+2]= ' ';
}


static void set_code_property_to (int p, boolx cp) 
{ 
  if ( cp == TRUE )
    set_code_property_to_Yes (p);
  else 
    if ( cp == FALSE )
      set_code_property_to_No (p);
}

static void quit (Widget w, XtPointer client_data, XtPointer call_data) 
{  
  XtDestroyApplicationContext(XtWidgetToApplicationContext(w));
  exit (0);
}

#endif


static void filter_word_not_code(char *word_not_code)
{
  int  nb_sep1, nb_sep2; /* number of separators in the 2 factorisations */
  int  i;
  int  cursor;   /* cursor to scan word_not_code */
  int  cursor_word, cursor_fac1, cursor_fac2; /* cursors to build new words */
  char letter; 
 
  i = 0;
  nb_sep1 = 0;
  nb_sep2 = 0;
  letter = word_not_code[i];
  while( letter != END_OF_STRING )
    {
      if ( letter == SEPARATOR1 )
	nb_sep1++;
      if ( letter == SEPARATOR2 )
	nb_sep2++;      
      letter = word_not_code[++i];
    }
  
  factorisation1 = (char *) malloc((i + 3 - nb_sep2 + nb_sep1) * sizeof(char));
  if( factorisation1 == NULL )
    free_resources(1);

  factorisation2 = (char *) malloc((i + 3 - nb_sep1 + nb_sep2) * sizeof(char));
  if( factorisation2 == NULL )
    free_resources(1);
  
  word = (char *) malloc((i - nb_sep1 - nb_sep2 + 1) * sizeof(char));
  if( word == NULL )
    free_resources(1);

  /* computation of factorisation1 and factorisation2 */  
  cursor = 0;
  cursor_word = 0;
  cursor_fac1 = 0;
  factorisation1[cursor_fac1++] = SEP1; /* the word begins with this separator */
  cursor_fac2 = 0;
  factorisation2[cursor_fac2++] = SEP1;
  letter = word_not_code[cursor];
  while( letter != END_OF_STRING )
    {
      if ( letter == SEPARATOR1 )
	{
	  factorisation1[cursor_fac1++] = SEP2;
	  factorisation1[cursor_fac1++] = SEP1;
	}
      else
	if ( letter == SEPARATOR2 )
	  {
	    factorisation2[cursor_fac2++] = SEP2;
	    factorisation2[cursor_fac2++] = SEP1;
	  }
	else
	  {
	    factorisation1[cursor_fac1++] = letter;
	    factorisation2[cursor_fac2++] = letter;
	    word[cursor_word++] = letter;
	  }
      letter = word_not_code[++cursor];
    }
  factorisation1[cursor_fac1++] = SEP2;
  factorisation1[cursor_fac1] = END_OF_STRING;
  factorisation2[cursor_fac2++] = SEP2;
  factorisation2[cursor_fac2] = END_OF_STRING;
  word[cursor_word] = END_OF_STRING;
}

