/* edmon.c 3.0 (KIEL) 11.94 */

#include "ext.h"

/* functions included:
 *
 *	static  void  setvars()
 *	static  void  monboard()
 *	static	void  initscrvars()
 *
 *		void  editmon()
 */


#define LINE1	"h: left  j: down  k: up  l: right  g: goto                  q: quit"
#define MLINE2  "r: show defining relations"
#define MLINE3  "'*' marks idempotent transformations"

static int scrcol,scrrow;	/* current screen position */
static int rowjump,coljump;     /* distance between columns and rows */
static int rowpage,colpage;     /* rows (columns) pro page */
static int rowmin,rowmax;	/* minimal and maximal row number */
static int colmin,colmax;	/* minimal and maximal column number */
static posint stateno,letterno;
	/* number of states and letters in the automaton */
static posint wordno;		/* number of words (monoid) */
static posint bufletter,bufstate;
	/* indices for the buffer to address letter/state transitions */
static posint bufword;		/* word index (monoid) */
static posint bc,br;
	/* bc is the first letter in the letter line and br is the first
 	 * state in the state line
 	 */
static boole table;
static monoid mon;
/******************************************************************/
/******************************************************************/
static void
setvars(dir,bcmax,brmax)
char dir;
posint bcmax, brmax;
/* change values of static variables according to the visible part
 * of the buffer 
 */
{
int row, col;
posint s;
	/* first: compute the value of the upper left corner (br,bc) */
	switch (dir) {
	case 'h': if (bc > colpage) bc -= colpage; else bc = 0;
		  break; 
	case 'j': br += rowpage;
		  if (br+rowpage > brmax) br = brmax-rowpage+1;
		  break;
	case 'k': if (br > rowpage) br -= rowpage; else br = 0;
		  break;
	case 'l': bc += colpage;
		  if (bc+colpage > bcmax) bc = bcmax-colpage+1;
		  break;
	case 'g': cleartut();
		  addstr("Goto element no: ");
		  gtyx(stdscr,row,col); refresh();
		  s = rdpnum(FALSE,row,col); s--;
		  if (s <= brmax) {
			  bc = 0; br = s;
			  if (br+rowpage > brmax) br = brmax - rowpage + 1;
			  bufstate = br; bufletter = bc;
		  } else { mvaddstr(LASTROW+1,LRCOL,"Not a valid number");
			   refresh(); sleep(1);
		  }
		  break;
	case 'i': /* this is the initialization  step */
		  bufletter = 0; bufstate = 0;
	          br = 0; bc = 0;
		  break;
	case 'u': /* refresh */
                  break;
        default : /* this should not happen */
		  errex(SWERR,"setvars");
	}
	bufstate = br + rowpage - 1;
	bufletter = bc + colpage - 1;
	/* second: compute the screen position of (bufstate,bufletter)
	 * e.g. the logical row for bufstate is the distance between
	 * the upperleft corner state and bufstate
	 * the pysical row (scrrow) is computed with a factor rowjump
	 * and the offset STARTROW
	 */
	scrrow = rowjump*(bufstate-br+1) + STARTROW;
	scrcol = coljump*(bufletter-bc+1) + STARTCOL;
} /* setvars */

/******************************************************************/
static void
monboard(dir)
char dir;
/* build-up the screen
 * dir is the direction indicator; dir == 'i' initialization
 */
{ int c,r,row,col;
  boole large=table&&(mon->mno>=TABLESWITCH);
  /* row and col store the logical row and column number; c and r contain
   * the physical (screen) row and column number
   */
	setvars(dir,stateno,wordno);
	clear();
	bufword = bufstate;
	bufstate = bufletter;

	/* print the word- and the stateline */
	if (table) maketitle(TRUE,"M U L T I P L I C A T I O N   T A B L E");
	else maketitle(TRUE,"S Y N T A C T I C   M O N O I D");
        move(INFOROW,STARTCOL);
        addstr("Number of elements: "); addstr(pi2a(mon->mno));
        addstr("       On page: "); addstr(pi2a(br+1));
	addstr(". - "); addstr(pi2a(br+rowpage)); addstr(".");
	if (!table) {
		addstr(" Number of states: ");
		addstr(pi2a(stateno+1));
	}

	c = (table && !bc)? colmin+coljump: colmin;
	col = (table && !bc)? 1: 0;
	if(large)
	  {for(;col<colpage;col++,c+=coljump)
		if (letterno <= ALPHSIZE) /* lowercase letters are symbolnames */
			mvaddch(STARTROW,c,itoc[mon->generator[bc+col]]);
		else {  mvaddch(STARTROW,c,'a');
			mvaddstr(STARTROW,c+1,pi2a(bc+col));
		}}
        else
	   for(;col<colpage;col++,c+=coljump)
		   mvaddstr(STARTROW,c,pi2a(bc+col));
	
	for(r = STARTROW+rowjump, row=0; row < rowpage; r+=rowjump,row++) {
		prtstr(r,STARTCOL,colmin-3,prword(br+row,mon,TRUE),1,STARTCOL+2,2,FALSE,FALSE);
		/* fill the row */
		for(c = colmin, col=0; col < colpage; c+=coljump, col++)
			mvaddstr(r,c,pi2a(mon->no2trans[br+row][bc+col]));
	}

	cleartut();
	mvaddstr(LASTROW+2,LRCOL,MLINE3);
	if (!table) 
           mvaddstr(LASTROW+1,LRCOL,MLINE2);
	mvaddstr(LASTROW,LRCOL,LINE1);
} /* monboard */

/******************************************************************/
static void
initscrvars()
/* initialize the screen variables */
{
	rowjump = STDROWJUMP;
	coljump = STDDCOLJUMP;
	wordno = mon->mno -1;
	stateno = mon->qno;
	/*
	 * xxxmin == startxxx + xxxjump
	 * xxxmax must be of the form startxxx+n*xxxjump
	 * xxxpage is the number of xxx's pro page
	 * but it is limited by the number of neccessary xxx's
	 */
	/* the length of the longest representative 
         * must be shorter than colmin 
         */
	colmin = (mon->sno > ALPHSIZE)? monmaxlen(mon)+6 : mon->no2length[mon->mno-1]+6;
	if(colmin > (2*COLS)/3) colmin=(2*COLS)/3;

	rowmin = STARTROW + rowjump;
	rowmax = ((LINES - ROWSUB)/rowjump)*rowjump;
	rowmax += (STARTROW - rowjump); 
	colmax = ((COLS - COLSUB)/coljump)*coljump;
	colmax += (STARTCOL - coljump);
	rowpage = (rowmax - rowmin)/rowjump + 1;
	colpage = (colmax - colmin)/coljump + 1;
	/* colmin has a value */
	colmin += STARTCOL;
	if (rowpage > wordno) rowpage = wordno+1;
	if (colpage > stateno) colpage = stateno+1;
} /* initscrvars */

/******************************************************************/
void editmon(mono,tab)
monoid mono;
boole tab;  /* tab == TRUE: multiplication table of a monoid */
/* invoke an editor to display the monoid mon */
{       
        mon = mono;
        table = tab;
	/* initialize the variables */
	initscrvars();

	/* build-up the screen */
	monboard('i');

	/* the editor-loop */
	for(;;) {
		refresh();		 /* show the new screen */

		switch(getch()) {
		case 'h': /* move left */
		case 'H': if (bufstate >= colpage) monboard('h');
			  break;
		case 'j': /* move down */
		case 'J': if (bufword < wordno) monboard('j');
			  break;
		case 'k': /* move up */
		case 'K': if (bufword >= rowpage) monboard('k');
			  break;
		case 'l': /* move right */
		case 'L': if (bufstate < stateno) monboard('l');
			  break;
		case 'g': /* goto */
		case 'G': monboard('g');
			  break;
		case 'r': /* relations */
		case 'R': if (!table) {
                                if(!mon->relationcomputed)mon2rel(mon);
				editrel(mon);
			  	monboard('u');
			  }
			  break;
		case 'q': /* quit */
		case 'Q': return;
		default: /* wrong action, do nothing */;
		}
	}
} /* editmon */
