#ifndef FA_UTILS_H
#define FA_UTILS_H

/*
   Bordeaux : 1995
   */

#include "ext.h"
#include <stdarg.h>


/* This module provides utilitaries for the finite automatons DFAs and NFAs */
/* In the comments, when a DFA (NFA) is called <name>_dfa (<name>_nfa), the 
   corresponding language will be called <name> */


/* initialize the fields of a structure DFA X_dfa with the others parameters
   that must therefore be coherents */
void fa_utils_init_fields_dfa(dfa X_dfa, posint qno, posint init, posint sno, 
			      mrkfin final, ddelta delta, boole minimal);
/* initialize the fields of a structure NFA X_nfa with the others parameters
   that must therefore be coherents */
void fa_utils_init_fields_nfa(nfa X_nfa, posint qno, posint sno, mrkfin final, 
			      ndelta delta, boole minimal, boole is_eps);
/* free the temporary DFAs used in a function : uses <stdarg.h> 
   bnfa : if TRUE, arguments must be NFAs else DFAs
   nb_arg is the number of DFA to free, the others arguments must be DFAs */
void fa_utils_free_temporary_fa(boole bnfa, int nb_arg, ...);
/* duplicate a DFA X_dfa */
dfa  fa_utils_duplicate_dfa(dfa X_dfa);



/* "cut" a DFA at the state newfinal and give the prefix DFA corresponding :
   the result is a new DFA that recognizes the word corresponding to path of X_dfa
   that start at the initial state of X_dfa and end at the newfinal state :
   thus it gives a class of prefix words of the language X (of DFA X_dfa) 
   newfinal must be a state of X_dfa */
dfa  fa_utils_prefix_i_dfa(dfa X_dfa, posint newfinal);
/* "cut" a DFA at the state newinit and give the suffix DFA corresponding :
   the result is a new DFA that recognizes the word corresponding to path of X_dfa
   that start at the newinit state and end at a final state of X_dfa :
   thus give a class of suffix words of the language X 
   newinit must be a state of X_dfa */
dfa  fa_utils_suffix_i_dfa(dfa X_dfa, posint newinit);



/* give the DFA corresponding at the empty language with an alphabet of size_al 
   letters */
dfa  fa_utils_empty_dfa(posint size_al);  
/* give the NFA corresponding at the empty language with an alphabet of size_al 
   letters */
nfa  fa_utils_empty_nfa(posint size_al);  

/* give the DFA corresponding at the empty word with an alphabet of size_al 
   letters */
dfa  fa_utils_eps_dfa(posint size_al);  
/* give the NFA corresponding at the empty word with an alphabet of size_al 
   letters */
nfa  fa_utils_eps_nfa(posint size_al);  

/* give the DFA corresponding at a language of one letter on an alphabet of 
   size_al letters */
dfa  fa_utils_one_letter_dfa(posint size_al, posint letterno);
/* give the NFA corresponding at a language of one letter on an alphabet of 
   size_al letters */
nfa  fa_utils_one_letter_nfa(posint size_al, posint letterno);

/* give the DFA corresponding at an alphabet of size_al letters */
dfa  fa_utils_alphabet_dfa(posint size_al);
/* give the NFA corresponding at an alphabet of size_al letters */
nfa  fa_utils_alphabet_nfa(posint size_al);



/* test if a (temporary) DFA corresponds to the empty language.
   The empty_full_lan()[win_one.c] function works only on the current language 
   assume that X_dfa is minimal */
boole fa_utils_test_empty_dfa(dfa X_dfa);
/* test if a (temporary) NFA corresponds to the empty language.
   The empty_full_lan()[win_one.c] function works only on the current language 
   assume that X_nfa is minimal */
boole fa_utils_test_empty_nfa(nfa X_nfa);

/* test if a (temporary) DFA corresponds to the empty word 
   assume that X_dfa is minimal */
boole fa_utils_test_eps_dfa(dfa X_dfa); 

/* test if epsilon is recognized by a DFA */
boole fa_utils_test_if_eps_in_dfa(dfa X_dfa);
/* test if epsilon is recognized by a NFA , assume that X_nfa has no
   epsilon transition */
boole fa_utils_test_if_eps_in_nfa(nfa X_nfa);

/* Verify if the language (of automatons X_dfa and X_nfa) is a sub-monoid :
   X_dfa must be minimal */
boole fa_utils_test_sub_monoid(dfa X_dfa, nfa X_nfa);


/* extract the well state of a DFA : if there is no well state, the result
   is (qno + 1) where qno is the max number of a state of X_dfa :
   X_dfa must be minimal, else it could have several well states */
posint fa_utils_well(dfa X_dfa); 

#endif 

 


