/* input.c 3.0 (KIEL) 11.94 */

#include "ext.h"

/* functions included:
 * 
 * 	static	boolx getyesno()
 *	static	void  callletters()
 *	static	void  callstates()
 *	static  boole getdata()
 *	static	boole dfainp()
 *	static	boole nfainp()
 *	static	boole enfainp()
 *	static	void  rexinp()
 *
 *		boole readexp()
 *		boole readaut()
 */


static boole ***dumnbuf = NULL;	/* dummy buffer for nfa's */
static posint **dumdbuf = NULL; /* dummy buffer for dfa's */

static posint ms,mq;		/* #letter, #states of mlang */

/******************************************************************/
/******************************************************************/
static boolx
getyesno(row,col,quittoo)
int row,col;
boole quittoo;
/* gets answer from (row,col) and prints "No" or "Yes"
 * quittoo TRUE iff q is also allowed
 * RETRUN: FALSE  iff no
 *	   TRUE   iff yes
 *         ISQUIT iff quit
 */
{	move(row,col); refresh();
	switch(getch()) {
	case 'n':
	case 'N': addstr("NO"); return FALSE;
	case 'q':
	case 'Q': if (quittoo) return ISQUIT;
	default : addstr("YES"); return TRUE;
	}
} /* getyesno */

/******************************************************************/
static void
callletters(s)
posint s;
/* assigns names to the letters of the alphabet
 * s is the size of the alphabet
 */
{	if (s <= ALPHSIZE) { /* use letters */
		if (s < 5)
			switch (s) {  /* write all letters */
			case 1: mvaddstr(ALPROW+1,IALPCOL,"a"); break;
			case 2: mvaddstr(ALPROW+1,IALPCOL,"a,b"); break;
			case 3: mvaddstr(ALPROW+1,IALPCOL,"a,b,c"); break;
			case 4: mvaddstr(ALPROW+1,IALPCOL,"a,b,c,d"); break;
			}
		else { /* use three dots for abbreviation */
		       mvaddstr(ALPROW+1,IALPCOL,"a,...,");
		       mvaddch(ALPROW+1,IALPCOL+6,itoc[s]);
		}
	} else { /* use a<number> for letters */
		mvaddstr(ALPROW+1,IALPCOL,"a1,...,a");
		mvaddstr(ALPROW+1,IALPCOL+8,pi2a(s));
	}
	refresh();
} /* callletters */

/******************************************************************/
static void
callstates(q)
posint q;
/* assigns names to the automaton states
 * q is the number of states
 */
{	if (q < 4) /* write all states */
		switch (q) {
		case 0: mvaddstr(QNOROW+1,IQNOCOL,"q0"); break;
		case 1: mvaddstr(QNOROW+1,IQNOCOL,"q0,q1"); break;
		case 2: mvaddstr(QNOROW+1,IQNOCOL,"q0,q1,q2"); break;
		case 3: mvaddstr(QNOROW+1,IQNOCOL,"q0,q1,q2,q3"); break;
		}
	else {  /* use three dots */
		mvaddstr(QNOROW+1,IQNOCOL,"q0,...,q");
  		mvaddstr(QNOROW+1,IQNOCOL+8,pi2a(q));
	}
  	refresh();
} /* callstates */

/******************************************************************/
static boolx
getdata(modify,size,name,ref,s,q)
boole modify;
posint size;
char *name,*ref;
posint *s,*q;
/* reads the data for an automaton
 * modify TRUE iff a automaton is modified
 * size stateno of m*fa
 * name is the name of the new language
 * ref is the reference
 * s,q  alphabet and stateset sizes
 * RETURN: TRUE iff ok; FALSE iff quit
 */
 { boolx quit;
		do {    if (modify) {
				drawmenu(MAUTMENU);
				mvaddstr(MODROW+3,IQNOCOL,pi2a(size+1));
			} else drawmenu(AUTMENU);
		rdnam(NAMROW,INAMCOL,name,CONROW,CONCOL);
		rdref(REFROW,IREFCOL,ref,CONROW,CONCOL);
		*s = rdpnum(FALSE,ALPROW,IALPCOL);
		callletters(*s);
		*q = rdpnum(FALSE,QNOROW,IQNOCOL);
		callstates((*q) - 1);
		mvaddstr(CONROW,CONCOL,"These data are correct? (y/n/q)");
	} while (!(quit=getyesno(CONROW,ICONCOL,TRUE)));
	return (quit == ISQUIT)? FALSE : TRUE;
 } /* getdata */

/******************************************************************/
static boole
dfainp(modify,org)
boole modify,org;
/* input a deterministic finite automaton
 * modify == TRUE iff a known language is modified
 * org TRUE iff the input is modified
 * RETURN: TRUE iff succesfully defined
 */
{char	name[NLEN+2];	/* name of new language */
 char   ref[REFLEN+2];  /* reference */
 posint s=0, n=0;	/* s,n: #letters,#states */
 posint init;		/* number of initial state */
 posint i=0,j;
 boole ret=FALSE;	/* return value */
 dfa curdfa,mdfa;	/* abbr. for curlan->ldfa and mlang->?dfa */

	if (modify) {
		/* assign mdfa the desired value */
		if ((org) && (mlang->inputis == ODFAINP))
			mdfa = mlang->odfa;
		else    mdfa = mlang->ldfa;
		i = mdfa->qno;
	}

	/* read data, and iff successful assign std-values to curlan */
	if (getdata(modify,i,name,ref,&s,&n)) {
		init = n;
		curlan = newlang(maxidx);
		for(i=0; i<NLEN; i++)
			curlan->name[i] = name[i];
		for(i=0; i<REFLEN; i++)
			curlan->ref[i] = ref[i];
		curlan->inputis = DFAINP;
		curlan->sizeal = s; 
		curlan->ldfa = newdfa();
		curdfa = curlan->ldfa;
		curdfa->sno = s;
		curdfa->qno = n-1;
                /* one more state, in case automaton is incomplete */
		curdfa->final = newfinal(curdfa->qno+1);     
		curdfa->delta = newddelta(s,curdfa->qno+1); 
		for(i=0; i <= n; i++) for(j=1; j <= s;)
			curdfa->delta[j++][i] = n;
		curdfa->minimal = FALSE;
                n--;

		/* if modify == TRUE, fill curdfa->delta and curdfa->final */
		if (modify) {
			ms = (s > mdfa->sno)? mdfa->sno : s;
			mq = (n > mdfa->qno)? mdfa->qno : n;
			if (mdfa->init <= mq) init = mdfa->init;
			for(i=0; i<=mq; i++) {
				curdfa->final[i] = mdfa->final[i];
				for(j=1; j <= ms; j++)
				  if (mdfa->delta[j][i] <= mq)
					curdfa->delta[j][i] = mdfa->delta[j][i];
		}	}

		curdfa->init  = edit(curdfa->delta,dumnbuf,curdfa->final,n,s,1,init,FALSE,FALSE);

		/* check if automaton is completely specified 
                 * if the automaton is not completely specified then
                 * add a sink-state
                 */
		n++;
		for(i=1; i<=s; i++) for(j=0; j<n; j++) 
			if (curdfa->delta[i][j] == n) {
					curdfa->qno = n;
					i = s; break;  /* quit loops */
				}
			
		ret = TRUE;
	}
	curlan->bdfa = TRUE;
	return ret;
} /* dfainp */

/******************************************************************/
static boole
nfainp(modify,org)
boole modify,org;
/* input a nondeterministic finite automaton
 * modify == TRUE iff a known language is modified
 * org TRUE iff the input is modified
 * RETURN: TRUE iff succesfully defined
 */
{char	name[NLEN+2];	/* name of new language */
 char   ref[REFLEN+2];  /* reference */
 posint s=0, n=0;	/* s,n: #letters,#states */
 posint i=0,j,k;
 boole ret=FALSE;	/* return value */
 nfa curnfa,mnfa;
 if(modify) 
   {if ((org) && (mlang->inputis == ONFAINP)) mnfa = mlang->onfa;
    else    mnfa = mlang->lnfa;
    i = mnfa->qno;
   }
 /* read data, and iff successful assign std-values to curlan */
 if (getdata(modify,i,name,ref,&s,&n)) {
		curlan = newlang(maxidx);
		for(i=0; i<NLEN; i++)
			curlan->name[i] = name[i];
		for(i=0; i<REFLEN; i++)
			curlan->ref[i] = ref[i];
		curlan->inputis = NFAINP;
		curlan->sizeal = s;
		curlan->lnfa = newnfa();
		curnfa = curlan->lnfa;
		curnfa->sno = s;
		curnfa->qno = --n;
		curnfa->infin = newfinal(curnfa->qno);
		curnfa->delta = newndelta(curnfa->sno,curnfa->qno);
		curnfa->is_eps = FALSE;
		curnfa->minimal = FALSE;

		/* if modify == TRUE, fill curnfa->delta and curnfa->infin */
		if(modify) 
                  {ms = (s > mnfa->sno)? mnfa->sno : s;
                   mq = (n > mnfa->qno)? mnfa->qno : n;
	  	   /* now mq and ms are the sno and qno for copy */
		   for(i=0; i<=mq; i++) 
                     {curnfa->infin[i] = mnfa->infin[i];
		      for(j=1; j <= ms; j++) for(k=0; k<=mq; k++)
		        cpdelta(curnfa->delta,j,i,k,mnfa->delta,j,i,k);
		     }
                  }

		(void)edit(dumdbuf,curnfa->delta,curnfa->infin,n,s,1,0,FALSE,TRUE);
		ret = TRUE;
	}
	curlan->bnfa = TRUE;
	return ret;
} /* nfainp */

/******************************************************************/
static boole
enfainp(modify)
boole modify;
/* input an epsilon finite automaton
 * modify == TRUE iff a known language is modified
 * RETURN: TRUE iff succesfully defined
 */
{char	name[NLEN+2];	/* name of new language */
 char   ref[REFLEN+2];  /* reference */
 posint s=0, n=0;	/* s,n: #letters,#states */
 posint start;          /* start=0 modify enfa start=1 modify nfa */
 posint i=0,j,k;
 boole ret=FALSE;	/* return value */
 nfa cureps,meps;

	if (modify) {
		if (mlang->inputis == OENFAINP)
                      { meps = mlang->oenfa; start=0; }
                else
                      { meps = mlang->lnfa; start=1; }
		i = meps->qno;
	}

	/* read data, and iff successful assign std-values to curlan */
	if (getdata(modify,i,name,ref,&s,&n)) {
		curlan = newlang(maxidx);
		for(i=0; i<NLEN; i++)
			curlan->name[i] = name[i];
		for(i=0; i<REFLEN; i++)
			curlan->ref[i] = ref[i];
		curlan->inputis = OENFAINP;
		curlan->sizeal = s;
		curlan->oenfa = newnfa();
		cureps = curlan->oenfa;
		cureps->sno = s;
		cureps->qno = --n;
		cureps->infin = newfinal(cureps->qno);
		cureps->delta = newendelta(cureps->sno,cureps->qno);
		cureps->is_eps = FALSE;
		cureps->minimal = FALSE;

		/* if modify == TRUE, fill cureps->delta and cureps->infin */
		if (modify) {
			ms = (s > meps->sno)? meps->sno : s;
			mq = (n > meps->qno)? meps->qno : n;
			/* now mq and ms are the sno and qno for copy */
			for(i=0; i<=mq; i++) {
				cureps->infin[i] = meps->infin[i];
				for(j=start; j <= ms; j++) for(k=0; k<=mq; k++)
				  cpdelta(cureps->delta,j,i,k,meps->delta,j,i,k);
		}	}

               (void)edit(dumdbuf,cureps->delta,cureps->infin,n,s,0,0,FALSE,TRUE);

		/* check if epsilon transitions exist */
		for(i=0; i <= n; i++) for(j=0; j <= n; j++)
			if (testcon(cureps->delta,0,i,j)) {
				cureps->is_eps = TRUE;
				i = n; break;      /* quit loops */
			}
		
		ret = TRUE;
	}
	return ret;
} /* enfainp */

/******************************************************************/
static void 
insertdot(regexp)
regex regexp;
/* transforms the regular expression in string to a regular expression with dots */
{
 char ch;
 posint length;
 posint newlength=0;
 posint j,len=0;
 boole finish=FALSE;
 string str=regexp->rex;
 string modstr;
 length=regexp->rexl;
 for(j=0; j < length ; )
    {
     ch = str[j++];
     if (((ch >= 'A') && (ch <= MAXLET)) || ((ch >= 'a') && (ch <= 'z')) || (ch == epsch))
        {
         if(finish)newlength++;
         else finish=TRUE;
        }
      else 
	switch(ch)
            {
             case complch:
             case '(': if(finish)
                         {newlength++;finish=FALSE;}
                       break;
             case concatch: 
             case minusch: 
             case insecch:
             case unionch: finish=FALSE;
                           break;
             case ')': finish=TRUE;
                       break;
             default:;
            }
    }
 if(newlength)
    {
     finish=FALSE;
     len=length+newlength;
     modstr=newstring(len+1);
     modstr[len]='\0';
     newlength=0;
     for(j=0; j < length ; )
       {
        ch = str[j++];
        if (((ch >= 'A') && (ch <= MAXLET)) || ((ch >= 'a') && (ch <= 'z')) || (ch == epsch))
           {
            if(finish)modstr[newlength++]=concatch;
            else finish=TRUE;
           }
         else 
	   switch(ch)
               {
                case complch:
                case '(': if(finish)
                            {modstr[newlength++]=concatch;finish=FALSE;}
                          break;
                case concatch: 
                case minusch: 
                case insecch:
                case unionch: finish=FALSE;
                              break;
                case ')': finish=TRUE;
                          break;
                default:;
               }
        modstr[newlength++]=ch;
       }
     dispose(regexp->rex);
     regexp->rex=modstr;
     regexp->rexl=newlength;
    }
} /* insertdot */

/******************************************************************/
static boole
rexinp(modify,gen)
boole modify;
boole gen;  /* gen == TRUE: input is generalized exp. */
{ boolx quit;
  posint s;
  char name[NLEN+2];
  char ref[REFLEN+2];
  int i;
  regex currex, mlpt;
  boole ret = FALSE;
	do {    if (modify) drawmenu(MREXMENU);
		else drawmenu(REXMENU);
		rdnam(NAMROW,INAMCOL,name,CONROW,CONCOL);
		rdref(REFROW,IREFCOL,ref,CONROW,CONCOL);
		if(modify) { 
                            s = mlang->sizeal;
                            mvaddstr(ALPROW,IALPCOL,pi2a(s));
                           }
                else s = rdpnum(FALSE,ALPROW,IALPCOL);
		callletters(s);
		mvaddstr(CONROW,CONCOL,"These data are correct? (y/n/q)");
	} while (!(quit=getyesno(CONROW,ICONCOL,TRUE)));

	/* read data, and iff successful assign std-values to curlan */
	if(quit != ISQUIT) 
           {
            curlan = newlang(maxidx);
            for(i=0; i<NLEN; i++) curlan->name[i] = name[i];
            for(i=0; i<REFLEN; i++) curlan->ref[i] = ref[i];
            curlan->inputis = REXINP;
            curlan->sizeal = s;
            curlan->lrex = newrex();
            currex = curlan->lrex;
            currex->sno = s;
            currex->grex = gen;
            currex->rexl = 0;
            currex->erexl = 0;
            currex->useda = -1;
            if(modify)    /* if modify == TRUE, fill currex*/
              { 
               if ((mlang->inputis == OREXINP) && gen) mlpt=mlang->orex;
               else mlpt = mlang->lrex;
               currex->rexl = mlpt->rexl;
               currex->rex = newrexstr(currex->rexl*3);
               (void)strcpy(currex->rex,mlpt->rex);
               currex->useda = mlpt->useda;
               for(i=0; i<=currex->useda; i++) 
                 {
                  currex->abbl[i] = mlpt->abbl[i];
                  currex->abbr[i] = newrexstr(currex->abbl[i]*3);
                  (void)strcpy(currex->abbr[i],mlpt->abbr[i]);
	         }
              }

		edrex(currex,TRUE);
		ret = TRUE;
	        curlan->brex = TRUE;
	   }
	return ret;
} /* rexinp */

/******************************************************************/
/******************************************************************/
/******************************************************************/
boole
readexp(modify,curchar)
boole modify;
char curchar;
/* do action on regular expression
 * modify TRUE iff mlang is known
 * curchar spezifies action 
 */
{ boole res=FALSE; /* TRUE iff language defined succesfully */
	switch(curchar) {
	case '1': /* regular expression */
		  if (!modify) 
		     help(FALSE,"Wrong choice: aborted");
                  else
                      {
                       if((!mlang->brex) || (mlang->lrex->grex)) comprex();
		       res = rexinp(modify,FALSE);
                      }
		  break;
        case '2': /* generalized regular expression */
		  if ((modify) && (!mlang->brex)) comprex();
		  res = rexinp(modify,TRUE);
		  break;
	default:  /* wrong choice */
		  help(FALSE,"Wrong choice: aborted");
	} /* end switch */
	return res;
} /* readexp */

/******************************************************************/
boole
readaut(modify,curchar)
boole modify;
char curchar;
/* do action on automaton
 * modify TRUE iff mlang is known
 * curchar spezifies action 
 */
{ boole res = FALSE; /* TRUE iff language defined succesfully */
	switch(curchar) {
	case '1': /* det. finite automaton */
		  if ((modify) && (!mlang->bdfa)) compdfa();
		  if (res = dfainp(modify,TRUE)) {
			curlan->inputis = DFAINP;
		  } break;
	case '2': /* nondet. finite automaton */
		  if ((modify) && (!mlang->bnfa)) compnfa();
		  res = nfainp(modify,TRUE);
		  break;
	case '3': /* epsilon finite automaton */
		  if ((modify) && (!mlang->inputis == OENFAINP)) compnfa();
		  res = enfainp(modify);
		  break;
	case '4': /* min. det. finite automaton */
		  if (modify) {
			compmdfa();
		  	res = dfainp(modify,FALSE);
			break;
		  }
	case '5': /* red. nondet. finite automaton */
		  if (modify) {
			compmnfa();
		  	res = nfainp(modify,FALSE);
			break;
		  }
	default:  /* wrong choice */
		  help(FALSE,"AUTOMATA: wrong choice; quit");
	} /* end switch */
	return res;
} /* readaut */
