/* pgcurses.c 3.0 (KIEL) 11.94 */

#include "ext.h"

/* functions included:
 *
 *	static string i2a()
 *	       boole startprint()
 *	       boole endprint()
 *	       int  getpgrow()
 *	       void pgclear()
 *	       void gohorizontal()
 *	       void govertical()
 *	       void pgaddch()
 *	       void pgmove()
 *	       void pgaddstr()
 *	       void pgmvaddstr()
 *	       void pgmvaddch()
 *	       void pggetyx()
 *	       void pghead()
 *	       void pgbreak()
 */
static char *name, *ref;
static char fname[PATHLEN+NLEN+5];
static char **page;
static int pageno,hpageno,vpageno;
static int pgrow, pgcol;

static boole printing = FALSE;

/******************************************************************/
static string
i2a(i)
int i;
{	(void)sprintf(dummy,"%d",i);
	return dummy;
} /* i2a */

/******************************************************************/
int
getpgrow()
{ return pgrow;
} /* getpgrow */

/******************************************************************/
boole
startprint(n,r,des)
char *n,*r,*des;
{ int i,j;

	if (printing) return FALSE;
	name = n; ref =r;

	/* first step: open file */
	(void)strcpy(fname,printdir);
	(void)strcat(fname,name);
	(void)strcat(fname,des);
	if ((amrp=fopen(fname,"w")) == NULL) {
#ifdef DEBUG
		perror("startprint 1");
#endif
		return FALSE;
	}

	printing = TRUE;
	help(FALSE,"PRINT: preparing file for printer ...");

	/* second step: alloc memory for page */
	page = (char **)calloc((posint)plines,sizeof(char *));
	if (page == NULL) {
#ifdef DEBUG
		perror("startprint 2");
#endif
		(void)fclose(amrp);
		printing = FALSE;
		buffree();freebuf(); (void)longjmp(env,1);
	}
	for(i=0; i < plines; i) {
		page[i] = (char *)calloc((posint)pcols,sizeof(char));
		if (page[i++] == NULL) {
#ifdef DEBUG
			perror("startprint 3");
#endif
			(void)fclose(amrp);
			for(j=i-1; --j;) dispose(page[j]);
			dispose(page);
			printing = FALSE;
			buffree();freebuf(); (void)longjmp(env,1);
	}	}

	/* clear page */
	for(i=0; i< plines; i++) for (j=0; j< pcols;)
		page[i][j++] = ' ';

	/* third step: initialize the running head */
	pageno = 1; hpageno = 1; vpageno = 1;

	/* last step: set page-var's */
	pgrow = 2; pgcol = 0;
	plines -= 3;

	return TRUE;
} /* startprint */

/******************************************************************/
boole
endprint()
/* free mem of page and do finishing touch */
{ int i;
	for(i=0; i< plines;)	dispose(page[i++]);
	dispose(page);

	if (ferror(amrp)) {
		(void)fclose(amrp);
		return FALSE;
	}
	(void)fclose(amrp);
	if (chmod(fname,0664) == -1) {
#ifdef DEBUG
		perror("endprint 1");
#endif
		return FALSE;
	}

	printing = FALSE;
	plines += 3;
	help(FALSE,"PRINT: preparing file for printer ... Done!");
	return TRUE;
} /* endprint */

/******************************************************************/
void
pgclear()
{ int i,j;
	for(i=0; i< plines+2; i++) for (j=0; j< pcols;)
		page[i][j++] = ' ';
} /* pgclear */

/******************************************************************/
void
gohorizontal()
{
	hpageno++;
} /* gohorizontal */

/******************************************************************/
void
govertical()
{
	hpageno = 1; vpageno++;
} /* govertical */

/******************************************************************/
void
pgaddch(ch)
char ch;
{
	page[pgrow][pgcol++] = ch;
} /* pageaddch */

/******************************************************************/
void
pgmove(row,col)
int row,col;
{
	pgrow = row+2, pgcol = col;
} /* pgmove */

/******************************************************************/
void 
pgaddstr(str)
string str;
{ int j;
	for(j=0; str[j] != '\0';)
		page[pgrow][pgcol++] = str[j++];
} /* pgaddstr */

/******************************************************************/
void
pgmvaddstr(row,col,str)
int row,col;
string str;
{ int j;
	pgrow = row+2; pgcol = col;
	for(j=0; str[j] != '\0';)
		page[pgrow][pgcol++] = str[j++];
} /* pgmvaddstr */

/******************************************************************/
void
pgmvaddch(row,col,ch)
int row,col;
char ch;
{
	pgrow = row+2; pgcol = col;
	page[pgrow][pgcol++] = ch;
} /* pgmvaddch */

/******************************************************************/
void
pggetyx(row,col)
int *row, *col;
{
	*row = pgrow-2;
	*col = pgcol;
} /* pggetyx */

/******************************************************************/
void
pghead()
{
	pgrow = 0; pgcol = 1;
	pgaddstr("AMORE  Name: "); 
	pgaddstr(name); pgaddstr("  --> ");
	pgaddstr(ref);
	pgrow = 1; pgcol = pcols - 50;
	pgaddstr(" Page "); pgaddstr(i2a(pageno));
	pgaddstr("    PID ("); pgaddstr(i2a(vpageno));
	pgaddch(','); pgaddstr(i2a(hpageno));
	pgaddch(')');
} /* pghead */

/******************************************************************/
void
pgbreak()
{ int i;
	/* close page */
	pghead();
	/* print page */
	for(i=0; i < plines+2; i++) {
		fputs(page[i],amrp);
		fputs("\n",amrp);
	}
	/* page break for printer device */
	fputs("\f\n",amrp);
	pageno++;
	/* clear page */
	pgclear();
} /* pgbreak */
