/*
   Bordeaux : 1995
   */


#include "ext.h"
#include "extbis.h"

#include "propcode.h"
#include "codetest.h"
#include "two_fact.h"
#include "edcode.h"
#include <string.h>
#include <ctype.h>

#define ACCEPT 'y'
#define REFUSE 'n'


/* Computes all code-properties for L -- assuming that L is a code */
static void code_properties (language L);

/* aks a question to the amore user and wait an answer */
static char choice(char *txt);

/* This procedure affects the code properties result to the language L if 
   this properties are not UN_KNOWN */
static void affect_properties(language L, boolx p_inf, boolx s_inf, boolx outfix, 
			      boolx hypercode, boolx uniform, boolx rs, boolx ls, 
			      boole fu);

/* ---------------------------- Global functions ------------------------------- */

void propcode_all_codes(language L)
/* assume that L->ldfa and L->lnfa are computed */
{
  boole test; 
  boole btmp;
  char  key;
  boole display = TRUE;

  if( L->code == UN_KNOWN ) /* this is the first time we go in this procedure */
    {
      help(FALSE,"Test if the language is a code");
      test = codetest_code(L->ldfa);
      L->code = test;
      L->saved = FALSE;
      updatewin(OWIN); /* to see if the language is a code or no */
    }
  else
    test = L->code;  

  btmp = ( test == TRUE && (L->prefix_code == UN_KNOWN || 
			    L->suffix_code == UN_KNOWN || 
			    L->bifix_code == UN_KNOWN || 
			    L->p_infix_code == UN_KNOWN || 
			    L->s_infix_code == UN_KNOWN || 
			    L->infix_code == UN_KNOWN || 
			    L->outfix_code == UN_KNOWN || 
			    L->hypercode == UN_KNOWN || 
			    L->right_semaphore_code == UN_KNOWN || 
			    L->left_semaphore_code == UN_KNOWN ||
			    L->uniform_code == UN_KNOWN || 
			    L->full_uniform_code == UN_KNOWN) );

  if ( btmp == TRUE ) /* L is a code and a propertie is not computed */
    code_properties(L);
  
  if ( test == FALSE && (strcmp(L->word_not_code, WORD_NOT_CODE_DEFAULT) == 0) )
    /* L->word_not_code was not computed */
    {
      if( L->prefix_code == UN_KNOWN ) /* first time we go in this case */
	{
	  L->prefix_code = FALSE;
	  L->suffix_code = FALSE;
	  L->bifix_code  = FALSE;
	  L->p_infix_code = FALSE;
	  L->s_infix_code = FALSE;
	  L->infix_code   = FALSE;
	  L->outfix_code  = FALSE;
	  L->hypercode    = FALSE;
	  L->right_semaphore_code = FALSE;
	  L->left_semaphore_code  = FALSE;
	  L->uniform_code = FALSE;
	  L->full_uniform_code = FALSE;

	  L->saved = FALSE;
	}
	  
      key = choice("Do you want Amore to compute a word with two factorisations? (y/n)");
      if ( key == ACCEPT ) /* compute word_not_code */
	{
	  two_fact_compute_word_not_code(L);
	  L->saved = FALSE;      
	}            
      else
	display = FALSE;
    }
  
  if( display == TRUE )
    {
#ifdef X_DISPLAY
      edcode_xdisplay_codes_info (L);
#else
      edcode_display_codes_info(L);
#endif
    }
}


/* ---------------------------- Static functions ------------------------------- */


static void code_properties (language L)
/* Here are the implications used to implement these tests on the codes 
   ( A ---> B <==> (if A then B) )
   
                       code <--------------------------------
                       /|\                                  |
                        |                                   |
        -----------------------------------                 |
       /|\                               /|\                |
        |                                 |                 |
  prefix code(1)                 (2)suffix code --------> stack code(13)
       /|\                               /|\
        |              (3)                |
        <---------- bifix code ----------->
       /|\             /|\               /|\
	|               <-----------------|---------- outfix code(11)
	|               |                 |             /|\
  p-infix code(4)       |        (5)s-infix code         |
       /|\              |                /|\             |
	|               |(6)              |              |
        <---------- infix code ----------->              |
       /|\             /|\               /|\             |
	|               <-----------------|---------- hypercode(12)
	|               |                 |             /|\
right semaphore(7)      |       (8)left semaphore        |
       /|\              |                /|\             |
	|               |(9)              |              |
	---------- full uniform -----------              |          
                        |                                |
			----------------------------> uniform(10)
*/
{
  char  key1;
  nfa   L_nfa;
  dfa   L_dfa;
  boole btmp;
  
  L_dfa = L->ldfa;
  L_nfa = L->lnfa;

  key1 = choice("Do you want individual confirmations for long computations ? (y/n)");

  /* computation of the prefix code property */
  if( L->prefix_code == UN_KNOWN )
    {
      help(FALSE,"Test if the language is a prefix code");
      L->prefix_code = codetest_prefix_code(L_dfa);
    }
  if ( L->prefix_code == FALSE )
    affect_properties(L, FALSE, UN_KNOWN, FALSE, FALSE, 
		      FALSE, FALSE, UN_KNOWN, FALSE );
  
  
  /* computation of the suffix code property */
  if( L->suffix_code == UN_KNOWN )
    {
      help(FALSE,"Test if the language is a suffix code");
      L->suffix_code = codetest_suffix_code(L_dfa);
    }
  if ( L->suffix_code == FALSE )
    affect_properties(L, UN_KNOWN, FALSE, FALSE, FALSE, 
		      FALSE, UN_KNOWN, FALSE, FALSE );
  

  /* computation of the bifix code property */
  L->bifix_code = codetest_bifix_code(L_dfa, L->prefix_code, L->suffix_code);


  /* computation of the p_infix code property */
  if( L->prefix_code == TRUE && L->p_infix_code == UN_KNOWN )
    if( key1 == REFUSE || 
       choice("Do you want the p-infix code property? (y/n)") == ACCEPT )
      {
	help(FALSE,"Test if the language is a p_infix code");
	L->p_infix_code = codetest_p_infix_code(L_nfa, L_dfa); 
      }

  /* update the properties depending of p-infix code property */
  if ( L->p_infix_code == FALSE )
    affect_properties(L ,FALSE , UN_KNOWN, UN_KNOWN, FALSE, 
		      FALSE, FALSE, UN_KNOWN, FALSE );


  /* computation of the s_infix code property */
  if( L->suffix_code == TRUE && L->s_infix_code == UN_KNOWN )
    if( key1 == REFUSE || 
       choice("Do you want the s-infix code property? (y/n)") == ACCEPT )
      {
	help(FALSE,"Test if the language is a s_infix code");
	L->s_infix_code = codetest_s_infix_code(L_nfa, L_dfa); 
      }

  /* update the properties depending of s-infix code property */
  if ( L->s_infix_code == FALSE )
    affect_properties(L, UN_KNOWN, FALSE , UN_KNOWN, FALSE,
		      FALSE, UN_KNOWN, FALSE, FALSE );

 
  /* computation of the infix code property */
  if( L->p_infix_code != UN_KNOWN && L->s_infix_code != UN_KNOWN )
    L->infix_code = codetest_infix_code( L_nfa, L_dfa,
					L->p_infix_code, L->s_infix_code);
  

  /* computation of the right semaphore code property */
  if( L->p_infix_code != FALSE && L->right_semaphore_code == UN_KNOWN )
    if( key1 == REFUSE || 
       choice("Do you want the right semaphore code property? (y/n)") == ACCEPT )
      {
	help(FALSE,"Test if the language is a right semaphore code");
	L->right_semaphore_code = codetest_right_semaphore_code(L_nfa, L_dfa, 
								L->prefix_code); 
      }

  /* update the properties depending of right semaphore code property */
  if( L->right_semaphore_code == FALSE )
    L->full_uniform_code = FALSE;
  else
    if( L->right_semaphore_code == TRUE && L->p_infix_code == UN_KNOWN )
      L->p_infix_code = TRUE;
  
  
  /* computation of the left semaphore code property */
  if( L->s_infix_code != FALSE && L->left_semaphore_code == UN_KNOWN )
    if( key1 == REFUSE || 
       choice("Do you want the left semaphore code property? (y/n)") == ACCEPT )
	{
	  help(FALSE,"Test if the language is a left semaphore code");
	  L->left_semaphore_code = codetest_left_semaphore_code(L_nfa, L_dfa, 
								L->suffix_code);
	}

  /* update the properties depending of left semaphore code property */
  if( L->left_semaphore_code == FALSE )
    L->full_uniform_code = FALSE;
  else
    if( L->left_semaphore_code == TRUE && L->s_infix_code == UN_KNOWN )
      L->s_infix_code = TRUE;

  /* update the properties depending of left semaphore and right semaphore code properties */
  if( L->infix_code == UN_KNOWN &&
     L->p_infix_code != UN_KNOWN && L->s_infix_code != UN_KNOWN )
    L->infix_code = codetest_infix_code(L_nfa, L_dfa, 
					L->p_infix_code, L->s_infix_code);

 
  /* computation of the  full uniform code property */
  btmp = (L->right_semaphore_code != FALSE && L->left_semaphore_code != FALSE && 
	  L->infix_code != FALSE );
  if ( btmp == TRUE && L->full_uniform_code == UN_KNOWN )
    {
      if( key1 == REFUSE || 
	 choice("Do you want the full uniform code property? (y/n)") == ACCEPT )
	{
	  help(FALSE,"Test if the language is a full uniform code");
	  L->full_uniform_code = codetest_full_uniform_code(L_dfa);
	}
    }
  else
    if( btmp == FALSE )
      L->full_uniform_code = FALSE;
  
  /* update the properties depending of full uniform code property */     
  if ( L->full_uniform_code == TRUE )
    affect_properties(L, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE, TRUE);
  /* computation of the uniform code property */
  if( L->hypercode != FALSE && L->uniform_code == UN_KNOWN )
    if( key1 == REFUSE || 
       choice("Do you want the uniform code property? (y/n)") == ACCEPT )
      {
	help(FALSE,"Test if the language is an uniform code");      
	L->uniform_code = codetest_uniform_code(L_dfa);	  
      }
  
  /* update the properties depending of uniform code property */
  if( L->uniform_code == TRUE )
    affect_properties(L, TRUE, TRUE, TRUE, TRUE, TRUE,
		      UN_KNOWN, UN_KNOWN, UN_KNOWN);
  else 
    if( L->uniform_code == FALSE )
      L->full_uniform_code = FALSE;


  /* computation of the outfix code property */
  if( L->bifix_code == TRUE && L->outfix_code == UN_KNOWN )
    if( key1 == REFUSE || 
       choice("Do you want the outfix code property? (y/n)") == ACCEPT )
      {
	help(FALSE,"Test if the language is an outfix code.");
	L->outfix_code = codetest_outfix_code(L_dfa);
      }

  /* update the properties depending of outfix code property */       
  if( L->outfix_code == FALSE )
    affect_properties(L, UN_KNOWN, UN_KNOWN, FALSE, FALSE, FALSE,
		      UN_KNOWN, UN_KNOWN, FALSE);


  /* computation of the hypercode property */
  if( L->infix_code != FALSE && 
     L->outfix_code != FALSE && L->hypercode == UN_KNOWN )
    if( key1 == REFUSE ||
       choice("Do you want the hypercode property? (y/n)") == ACCEPT )
      {
	help(FALSE,"Test if the language is hypercode.");
	L->hypercode = codetest_hypercode(L_dfa, L_nfa);
      }

  /* update the properties depending of hypercode property */
  if ( L->hypercode == TRUE )
    affect_properties(L, TRUE, TRUE, TRUE, TRUE, UN_KNOWN,
		      UN_KNOWN, UN_KNOWN, UN_KNOWN);
  else 
    if( L->hypercode == FALSE )
      affect_properties( L, UN_KNOWN, UN_KNOWN,
			UN_KNOWN, FALSE, FALSE, UN_KNOWN, UN_KNOWN, FALSE);

  
  /* computation of the stack code property */
  if( L->suffix_code != TRUE && L->stack_code == UN_KNOWN )
    {
      if( key1 == REFUSE ||
	 choice("Do you want the stack code property? (y/n)") == ACCEPT )
	{
	  help(FALSE,"Test if the language is a stack code");
	  L->stack_code = codetest_stack_code(L_dfa, L_nfa);
	}    
    }
  else
    if( L->suffix_code == TRUE )
      L->stack_code = TRUE;

  L->saved = FALSE;
}      



static char choice(char *txt)
{
  char result;

  help(FALSE, txt);
  result = tolower(getch());
  while( result != ACCEPT && result != REFUSE )
    result = tolower(getch());

  return result;
}


static void affect_properties(language L, boolx p_inf, boolx s_inf, boolx outfix, 
			      boolx hypercode, boolx uniform, boolx rs, boolx ls, 
			      boolx fu)
{
  if( p_inf != UN_KNOWN )
    L->p_infix_code = p_inf;

  if( s_inf != UN_KNOWN )
    L->s_infix_code = s_inf;
  
  if( p_inf != UN_KNOWN && s_inf != UN_KNOWN )    
    L->infix_code = p_inf && s_inf;
  else
    if( p_inf == FALSE || s_inf == FALSE )
      L->infix_code = FALSE;

  if( outfix != UN_KNOWN )
    L->outfix_code = outfix;

  if( hypercode != UN_KNOWN )
    L->hypercode = hypercode;

  if( uniform != UN_KNOWN )
    L->uniform_code = uniform;

  if( rs != UN_KNOWN )
    L->right_semaphore_code = rs;

  if( ls != UN_KNOWN )
    L->left_semaphore_code  = ls;

  if ( fu != UN_KNOWN )
    L->full_uniform_code = fu;
}

