/* rexnfa.c 3.0 (KIEL) 11.94 */

#include "ext.h"

/******************************************************************/
/* functions included:
 *	static boole complast()
 *	static boole compfirst()
 *	static void getexpr()
 *	static void compff()
 *		nfa rex2nfa()
 */
/******************************************************************/

#define skipch		'!'
#define unskipch 	'?'

static string rex,prerex;
/* postfix (ab.) and prefix (.ab) form of input string (a.b) */
static posint pi;
static posint constate;   /* state to which compfirst connects */
static posint length; /* length of expressions */

static posint *posch, *rposch; /* transformation table for letter -> state */
static posint *skipval;
static posint hlp;  /* store values, counter */
static posint sno; /* size of alphabet */

/* variables which run through rex and prerex */
static posint lastrex,lastprerex;
static posint firstprerex;
static posint getrex, getprerex;
static posint ffrex, ffprerex;

/******************************************************************/
/******************************************************************/
static boole
complast(buf)
mrkfin buf;
/* computes { a | A*a intersection L(rex) != emptyset } in buf
 * RETURN: TRUE iff the emptyword is in L(act_rex)
 *         static variable prerex: prefix-form of rex
 *	   static var rposch: posch for prerex
 * recursiv 
 * scans rex form left to right (lastrex), constructs prerex (lastprerex) and
 * rposch and marks last states in buf "final"
 *
 *		last(e1 union e2) = last(e1) union last(e2)
 *		last(e1 concat e2) = if (eps in e2) last(e1) union last(e2)
 *				     else last(e2)
 *		last(e star) = last(e)
 */
{ char op;
  boole epslbl1, epslbl2; /* return values for eps-property */
	tick();
	op = rex[lastrex--];
	while (isdigit(op)) {
		prerex[--lastprerex] = op;
		tick();
		op = rex[lastrex--];
	}
	switch(op) {
	case unionch: epslbl2 = complast(buf);
		      epslbl1 = complast(buf);
		      prerex[--lastprerex] = op;
		      return(epslbl1 || epslbl2);
	case plusch:  epslbl1 = complast(buf);
		      prerex[--lastprerex] = op;
		      return(epslbl1);
	case starch:  (void)complast(buf);
	case epsch:   prerex[--lastprerex] = op;
		      return(TRUE);
	case 'A':
	case 'a':
	case 'b':
	case 'c':
	case 'd':
	case 'e':
	case 'f':
	case 'g':
	case 'h':
	case 'i':
	case 'j':
	case 'k':
	case 'l':
	case 'm':
	case 'n':
	case 'o':
	case 'p':
	case 'q':
	case 'r':
	case 's':
	case 't':
	case 'u':
	case 'v':
	case 'w':
	case 'x':
	case 'y':
	case 'z':     if (buf != NULL) {
			setfinalT(buf[posch[lastrex+1]]);
		      }
		      rposch[lastprerex-1] = posch[lastrex+1];
	case langch:  prerex[--lastprerex] = op;
		      return(FALSE);
	case concatch: /* scan e2 first and if eps not in e2 skip over e1 */
		      if (complast(buf)) epslbl1 = complast(buf);
		      else { (void)complast((mrkfin)NULL); epslbl1 = FALSE; }
		      prerex[--lastprerex] = op;
		      return(epslbl1);
	default: errex(SWERR,"complast");
	}
	/* NOTREACHED */
} /* complast */

/******************************************************************/
static boole
compfirst(buf)
ndelta buf;
/* computes { a | aA* intersection L(rex) != emptyset } in buf
 * RETURN: TRUE iff the emptyword is in L(act_rex)
 * recursiv 
 * scans prerex from left to right (firstprerex) and connect the first states
 * to constate in buf
 *
 *		first(e1 union e2) = first(e1) union first(e2)
 *		first(e1 concat e2) = if (eps in e1) first(e1) union first(e2)
 *				      else first(e1)
 *		first(e star) = first(e)
 */
{ char op;
  boole epslbl1, epslbl2; /* return values for eps-property */
	switch(op = prerex[firstprerex++]) {
	case skipch:  firstprerex = skipval[firstprerex-1];
	case unskipch:return(compfirst(buf));
	case unionch: epslbl2 = compfirst(buf);
		      epslbl1 = compfirst(buf);
		      return(epslbl1 || epslbl2);
	case plusch:  epslbl1 = compfirst(buf);
		      return(epslbl1);
	case starch:  (void)compfirst(buf);
	case epsch:   return(TRUE);
	case 'A': if (buf != NULL)
			for(hlp=1; hlp <= sno; hlp++)
			   connect(buf,hlp,constate,rposch[firstprerex-1]);
		  return(FALSE);
	case 'a': op = prerex[firstprerex]; hlp = firstprerex -1; pi = 1;
		  if (('0' <= op) && (op <= '9')) {
			pi = 0;
		  	while(('0' <= op) && (op <= '9')) {
			 pi *= 10; pi += (op - '0');
			 op = prerex[++firstprerex];
		  }     }
		  if (buf != NULL) connect(buf,pi,constate,rposch[hlp]);
		  return(FALSE);
	case 'b':
	case 'c':
	case 'd':
	case 'e':
	case 'f':
	case 'g':
	case 'h':
	case 'i':
	case 'j':
	case 'k':
	case 'l':
	case 'm':
	case 'n':
	case 'o':
	case 'p':
	case 'q':
	case 'r':
	case 's':
	case 't':
	case 'u':
	case 'v':
	case 'w':
	case 'x':
	case 'y':
	case 'z':     if (buf != NULL)
			connect(buf,(op-'a'+1),constate,rposch[firstprerex-1]);
	case langch:  return(FALSE);
	case concatch:/* skip first(e2) iff eps not in e1 */
		      if (compfirst(buf)) return(compfirst(buf));
		      else { (void)compfirst((ndelta)NULL); return(FALSE); };
	}
	/* NOTREACHED */
} /* compfirst */

/******************************************************************/
static void
getexpr(copy)
/* gets an expression from rex (getrex) and
 * iff copy == TRUE : copies it to prerex (getprerex)
 * else             : count the number of symbols
 */
{ char op; /* act. symbol */
	switch (op = rex[getrex--]) {
	case unionch:
	case concatch: getexpr(copy);
	case plusch:
	case starch:   getexpr(copy);
		       if (copy) prerex[--getprerex] = op;
		       else length += 2;
		       break;
	case 'A':
	case 'a':
	case 'b':
	case 'c':
	case 'd':
	case 'e':
	case 'f':
	case 'g':
	case 'h':
	case 'i':
	case 'j':
	case 'k':
	case 'l':
	case 'm':
	case 'n':
	case 'o':
	case 'p':
	case 'q':
	case 'r':
	case 's':
	case 't':
	case 'u':
	case 'v':
	case 'w':
	case 'x':
	case 'y':
	case 'z': if (copy) {
			prerex[--getprerex] = op;
			rposch[getprerex] = posch[getrex+1];
		     } else length++;
		     break;
	case epsch:
	case langch:
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9': if (copy)
			prerex[--getprerex] = op;
		  else length++;
		  break;
	default: errex(SWERR,"getexpr");
	}
} /* getexpr */

/******************************************************************/
static void
compff(f,buf)
posint f;
ndelta buf;
/* compute the follow sets
 * scans rex from left to right (ffrex), and copies some information
 * to prerex (for compfirst) with ffprerex
 *		define X by
 *		X(langch,f) = X(epsch,f) = emptyset
 *		X(letter,f) = {letter}xfirst(f)
 *		X(e1 union e2,f) = X(e1,f) union X(e2,f)
 *		X(e1 con e2,f) = X(e1,e2 con f) union X(e2,f)
 *		X(e star,f) = X(e, e union f)
 *
 *		then
 *		X(e,f) = follow(e) union last(e)xfirst(f)
 *
 *	we have: X == compff
 *               e == rex
 *               f == prerex+ffprerex
 */
{ char ch;
  posint oldffprerex,  /* to store ffprerex */
	 skippos;      /* the last skip position */
	tick();
	ch = rex[ffrex--];
	while (isdigit(ch)) {
		prerex[--ffprerex] = ch;
		tick();
		ch = rex[ffrex--];
	}
	switch(ch) {
	case unionch:	/* call compff two times starting at f */
			prerex[--ffprerex] = skipch; skippos = ffprerex;
			skipval[skippos] = f;
			compff(f,buf);
			compff(f,buf);
			prerex[--ffprerex] = unionch;
			prerex[skippos] = unskipch;
			break;
	case concatch:  /* call compff once from f and then from ffprerex */
			prerex[--ffprerex] = skipch; skippos = ffprerex;
			skipval[skippos] = f;
			compff(f,buf);
			prerex[--ffprerex] = concatch; /* e2 concat f */
			oldffprerex = ffprerex;
			compff(ffprerex,buf);
			prerex[--ffprerex] = ch; 
			prerex[skippos] = unskipch;
			prerex[oldffprerex] = unskipch;
			break;
	case starch:
	case plusch:	/* skip multiple stars and pluses (but a star overwrites
			 * a plus)
			 */
			while ((rex[ffrex]==starch) || (rex[ffrex]==plusch)) {
				if (rex[ffrex--] == starch) ch = starch;
				tick();
			}
			prerex[--ffprerex] = skipch; skippos = ffprerex;
			skipval[skippos] = f;
			/* get the expression under the star */
			getrex = ffrex; getprerex = ffprerex;
			getexpr(TRUE);
			ffprerex = getprerex;
			prerex[--ffprerex] = unionch; /* e union f */
			/* notice the union and the following stuff must
			 * be deleted after X(e, e union f) has been
			 * computed (the union-sign must become a star or plus)
			 */
			oldffprerex = ffprerex;
			compff(ffprerex,buf);
			ffprerex = oldffprerex;
			prerex[ffprerex] = ch;
			prerex[skippos] = unskipch;
			break;
	case 'A':
	case 'a':
	case 'b':
	case 'c':
	case 'd':
	case 'e':
	case 'f':
	case 'g':
	case 'h':
	case 'i':
	case 'j':
	case 'k':
	case 'l':
	case 'm':
	case 'n':
	case 'o':
	case 'p':
	case 'q':
	case 'r':
	case 's':
	case 't':
	case 'u':
	case 'v':
	case 'w':
	case 'x':
	case 'y':
	case 'z':	constate = posch[ffrex+1];
			firstprerex = f;
			(void)compfirst(buf);
			prerex[--ffprerex] = ch;
			rposch[ffprerex] = constate;
			break;
	case epsch:
	case langch:	prerex[--ffprerex] = ch;
			break;
	default: errex(SWERR,"compff");
	}
} /* compff */
/******************************************************************/
/******************************************************************/
nfa rex2nfa(re)
register regex re;
/* computes a nondeterministic automaton from a given
 * regular expression (must not be generalize!, i.e. no
 * complement, intersection, ...) 
 * a "fast" algorithm is used:
 * 	G.Berry, R.Sethi
 *	From Regular Expressions to Deterministic Automata
 *	TCS 48(1986), pp. 117-126
 *      desciption:
 * RETURN: the NFA
 * 
 * always: state 0 is only final state !!! used in genrexnf.c !!!
 */
{ posint stateno, /* number of states in the NFA (one more than 
		   * #letters in the reg. expr.) */
	 prerexlen; /* length of prerex = length of rex + # stars in
		     * rex + # plus in rex + # concat in rex */
  nfa result;
  register posint i;
  posint concats=0; /* number of concats in re */

	/* init static vars */
	rex = re->exprex;
	posch = newarray(re->erexl);
	sno = re->sno;

	/* known values of the result */
	result = newnfa();
	result->sno = re->sno;
	result->is_eps = FALSE;
	result->minimal = FALSE;

	/* compute the number of states */
	stateno = 0; length = 0;
	for(i=0; i<re->erexl; i++) {
		tick();
		switch(rex[i]) {
		case 'A':
		case 'a':
		case 'b':
		case 'c':
		case 'd':
		case 'e':
		case 'f':
		case 'g':
		case 'h':
		case 'i':
		case 'j':
		case 'k':
		case 'l':
		case 'm':
		case 'n':
		case 'o':
		case 'p':
		case 'q':
		case 'r':
		case 's':
		case 't':
		case 'u':
		case 'v':
		case 'w':
		case 'x':
		case 'y':
		case 'z': posch[i] = ++stateno;
			  break;
		case plusch:
		case starch: getrex = i-1;
			     getexpr(FALSE);
			     while ((rex[i+1] == plusch) || (rex[i+1] == starch)) {
				i++; tick();
			     }
		case unionch: length++;break;
		case concatch: length++;concats++;break;
		default: length++;
	}	}

	/* now fill in the missing values */
	result->qno = stateno;
	result->infin = newfinal(result->qno);
	result->delta = newndelta(result->sno,result->qno);

	/* a special initial state (unique) */
	setinit(result->infin[0]);

	/* first step: the final states and prerex
	 * prepare prerex (length gt length of rex: we need
	 * some place to copy star-expressions more than once)
	 * define rposch and skipval
	 * complast returns TRUE: 
         *  the empty word belongs to the language 
         * the only initial state must marked final
	 * complast marks the last symbols final in 
	 * result->infin
	 */
	prerexlen = re->erexl+length+concats;
	prerex = newstring(prerexlen+1);
	prerex[prerexlen] = '\0'; prerex[re->rexl] = '\0';
	rposch = newarray(prerexlen+1);
	skipval = newarray(prerexlen+1);
	lastprerex = re->erexl; lastrex = lastprerex-1;
	setfinal(result->infin[0],complast(result->infin));

	/* second step: the start state
	 * compfirst connects all first symbols with constate
	 * an puts the connections in result->delta
	 */
	firstprerex = 0; constate = 0;
	(void)compfirst(result->delta);

	/* third step: the connections
	 * call compff with an epsilon string in prerex
	 */
	ffprerex = prerexlen-1; ffrex = re->erexl-1;
	prerex[ffprerex] = epsch;
	compff(ffprerex,result->delta);

	freebuf();
	return(result);
} /* rexnfa */
