/*
   Bordeaux : 1995
   */

#include "ext.h"
#include "extbis.h"

#include "two_fact.h"
#include "adjust.h"
#include "fa_utils.h"
#include "fa_word.h"
#include <stdlib.h>
#include <setjmp.h>

/* static variables for the research of multiples factorisations of a word */
static int length;             /* length of the word for wich we search two 
				  factorisations */
static int *factor1, *factor2; /* arrays with the length of each factor */
static int nb_facts;           /* number of factorisations found */
static dfa global_X_dfa;       /* DFA of the current language */
static posint global_well;     /* and its well sate */
static jmp_buf two_facts_found;

/* Recursive backtracked version of two_facts */
/* Parameters:
   int ord_fact;      number of factors found
   int read;          number of chars read in the full word
   fa_word w;         the remaining of the word to factorize */
static void recursive_two_facts (fa_word w, int read, int ord_fact);

/* Computes the double factorisation of a word of L -- assuming that L is 
   not a code */
static void two_facts (dfa X_dfa, fa_word word);

/* Give the DFA that recognizes the same language that X_dfa, but on an 
   alphabet of one more letter (X_dfa->sno + 1) */
static dfa dfa_more_one_letter(dfa X_dfa);

/* macro function to compute the length of one of the two factorisations
   factor1 and factor2 */
#define LEN_FAC(_FAC_, _LEN_) \
{\
   (_LEN_) = 0; \
   for ( cursor = (_FAC_) ; *cursor != 0 ; cursor++ ) \
     (_LEN_)++; \
   (_LEN_)--; \
}

/* ---------------------------- Global functions ------------------------------- */

void two_fact_compute_word_not_code(language L)
{
  if ( fa_utils_test_if_eps_in_dfa(L->ldfa) == TRUE )
    {
      L->word_not_code = (char *) realloc(L->word_not_code, 
					  (strlen(EPS_IN_DFA)+1) * sizeof(char));
      if( L->word_not_code == NULL )
	free_resources(1);
      strcpy(L->word_not_code, (char *) EPS_IN_DFA);	  
    }
  else
    {
      dfa      L_dfa;
      fa_word  word_tmp1, word_tmp2;
      char    *result;
      language old_language;

      /* we minimalize the DFA of L */
      old_language = langs[curidx];
      langs[curidx] = L;
      compmdfa();
      langs[curidx] = old_language;

      L_dfa = L->ldfa;

      help(FALSE,"Search a word with two factorisations");
      word_tmp1 = two_fact_find_a_word_with_double_factorisation(L_dfa);
      help(FALSE,"Compute the two factorisations");
      word_tmp2 = two_fact_fa_word_with_two_factorisations(word_tmp1, L_dfa);
      
      L->word_not_code = fa_word_convert_fa_word_to_string(word_tmp2, L->sizeal);
      
      free(word_tmp1);
      free(word_tmp2);
    }
}


fa_word two_fact_find_a_word_with_double_factorisation(dfa X_dfa)
/* Algorithm :
   X_dfa is a DFA for the language X
   1) we compute Y = (X.{$}.X+) Inter (Shuffle(X,{$}) - ({$}.X U X.{$})).X*
      where $ is a special character (the letter (X_dfa->sno + 1))
   2) we extract a word of Y and we remove $ of this word : 
      it has at least two factorisations */
{
  posint  newsno;
  dfa     new_X_dfa, special_char_dfa;
  dfa     plus_new_X_dfa, star_new_X_dfa;
  dfa     dfa1, dfa2, dfa3, dfa4, dfa5, dfa6, dfa7, dfa8;
  dfa     tmp_dfa;
  fa_word tmp_result, result;

  new_X_dfa = dfa_more_one_letter(X_dfa); 
  newsno    = new_X_dfa->sno;

  special_char_dfa = fa_utils_one_letter_dfa(newsno, newsno);

  plus_new_X_dfa = adjust_plus_dfa(new_X_dfa); 
  star_new_X_dfa = adjust_star_dfa(new_X_dfa);

  dfa1 = adjust_concat_dfa(new_X_dfa, special_char_dfa);  
  dfa2 = adjust_concat_dfa(special_char_dfa, new_X_dfa);
  dfa3 = adjust_concat_dfa(dfa1, plus_new_X_dfa);
  dfa4 = dfamdfa(adjust_shuffle_dfa(new_X_dfa, special_char_dfa), TRUE);
  dfa5 = adjust_union_dfa(dfa1, dfa2);
  fa_utils_free_temporary_fa(FALSE, 5, dfa1, dfa2, special_char_dfa, 
			     new_X_dfa, plus_new_X_dfa);

  dfa6 = dfamdfa(adjust_minus_dfa(dfa4, dfa5), TRUE);
  dfa7 = adjust_concat_dfa(dfa6, star_new_X_dfa);
  fa_utils_free_temporary_fa(FALSE, 4, dfa4, dfa5, dfa6, star_new_X_dfa);

  dfa8 = dfamdfa(adjust_insec_dfa(dfa3, dfa7), TRUE);
  fa_utils_free_temporary_fa(FALSE, 2, dfa3, dfa7);

  tmp_result = fa_word_find_a_word_in_dfa(dfa8, FALSE);
  result     = fa_word_remove_letter(tmp_result, newsno);

  freedfa(dfa8);
  free(tmp_result);
  return result;
}



fa_word two_fact_fa_word_with_two_factorisations(fa_word src, dfa X_dfa)
{
  int     i, l, i1, i2;
  int    *cursor;
  int     length_fac1, length_fac2;/*numbers of separators in each factorisation */
  int     length_word_not_code;    /* length of the final word word_not_code */
  int     cpt1, cpt2;              /* cursors "on the two factorisations" */
  fa_word w2;

  /* Building of the arrays factor1 and factor2 */
  two_facts(X_dfa, src);

  /* compute the number of separators for the first factorisation */
  LEN_FAC(factor1, length_fac1);

  /* compute the number of separators for the second factorisation */
  LEN_FAC(factor2, length_fac2);

  length_word_not_code = length + length_fac1 + length_fac2 + 1;
  w2= (fa_word) malloc(length_word_not_code * sizeof(int));
  if( w2 == NULL )
    free_resources(1);
  
  /* construction of word_not_code */
  i1 = 0;
  i2 = 0;
  l  = 0;
  cpt1 = factor1[0];
  cpt2 = factor2[0];

  for ( i = 0 ; i < length_word_not_code - 1 ; i++ )
    {
      if ( cpt1 == 0 )  /* mark a factor of the first factorisation */
	{
	  w2[i] = PI_SEPARATOR1;
	  cpt1 = factor1[++i1];
	}
      else
	if ( cpt2 == 0 ) /* mark a factor of the second factorisation */
	  {
	    w2[i] = PI_SEPARATOR2;
	    cpt2 = factor2[++i2];
	  }
      else
	{
	  w2[i] = src[l++];
	  cpt1--;
	  cpt2--;
	}
    }

  w2[i] = END_OF_FA_WORD;
  
  free(factor1);
  free(factor2);

  return w2;
}

/* ---------------------------- Static functions ------------------------------- */

static void recursive_two_facts(fa_word w, int read, int ord_fact)     
/* Assume we have found a factorisation till a certain position.
   If we are at the end of the word, then the full word has been factorized.
   Else, we parse the word until we find one factor and try to factorize the 
   rest of the word.
   If we cannot find a new factor, either because we are at the end of the word
   or have reached the well state, we backtrack and retry with a longer 
   factor.*/
{
  int    l;       /* number of chars read */
  posint q;       /* current state in the automaton */
  posint q0;      /* next state */

  l = 0;
  q = global_X_dfa->init;

  if ( read == length ) 
    {
      nb_facts++;
      if ( nb_facts == 1 )
	{
	  int i;
	  /* copy factor into factor2 and go for the second factorisation */
	  for ( i =  0 ; i <= length ; i++ ) 
	    factor2[i] = factor1[i];
	   return;
	}
      longjmp (two_facts_found, 2);
    }
  
  do 
    {
      q0 = global_X_dfa->delta[w[l++]][q];

      if (isfinal(global_X_dfa->final[q0])) 
	{
	  factor1[ord_fact] = l;
	  recursive_two_facts (&w[l], read+l, ord_fact+1);
	  factor1[ord_fact] = 0;
	}
      q= q0;
    } 
  while ( (read + l != length) && (q0 != global_well) );

}


static void two_facts(dfa X_dfa, fa_word word)
{
  int i;

  /* initializing global variables for recursive_two_facts */
  nb_facts = 0;
  length   = fa_word_len(word);

  factor1 = (int *) calloc(length+1, sizeof(int));
  if( factor1 == NULL )
    free_resources(1);
      
  factor2 = (int *) calloc(length+1, sizeof(int));
  if( factor2 == NULL )
    {
      free(factor1);
      free_resources(1);
    }

  global_X_dfa = X_dfa;
  global_well  = fa_utils_well(X_dfa);

  if( setjmp(two_facts_found) == 0 )
    recursive_two_facts (word, 0, 0);
}



static dfa dfa_more_one_letter(dfa X_dfa)
{
  posint sno;
  posint qno;
  posint newwell; /* well state of the new automaton */
  posint newqno;  /* number of states of the new automaton */
  posint newsno;  /* number of letters of the new alphabet */
  posint init;
  dfa    new_X_dfa;
  int    i,j;
  
  qno    = X_dfa->qno;
  init   = X_dfa->init;
  sno    = X_dfa->sno;
  newsno = sno + 1; 

  newwell = fa_utils_well(X_dfa);
  if ( newwell == qno + 1 )  
    /* X_dfa has no well state : we create this state in new_X_dfa */
    newqno = newwell;
  else
    newqno = qno;

  new_X_dfa = newdfa();
  
  fa_utils_init_fields_dfa(new_X_dfa, newqno, init, newsno, newfinal(newqno), 
			   newddelta(newsno, newqno), FALSE);
  
  for ( i = 0 ; i <= qno ; i++ )
    {
      for ( j = 1 ; j <= sno ; j++ )
	new_X_dfa->delta[j][i] = X_dfa->delta[j][i];
      new_X_dfa->delta[newsno][i] = newwell;
      new_X_dfa->final[i] = X_dfa->final[i];
    }

  if ( newqno != qno )
    {
      for ( j = 1 ; j <= newsno ; j++ )
	new_X_dfa->delta[j][newqno] = newqno;
      new_X_dfa->final[newqno] = FALSE;
    }

  return new_X_dfa;
}





    
