/*types.h 3.0 (KIEL) 11.94 */
/******************************************************************/

/* basic types */
typedef unsigned int    posint;       /* the basic type for numbers */
typedef char            boole;        /* boole==FALSE,TRUE */
typedef char            boolx;        /* boolx==FALSE,TRUE, UN_KNOWN,... */
typedef char*           string;
typedef string*         arrayofstring;
typedef posint*         array;
typedef array*          arrayofarray;
typedef boole*          b_array;
typedef b_array*        arrayofb_array;

/* types for deterministic finite automaton */

typedef boolx   *mrkfin;        /* for final states (FINMARK iff final)
                                 * type boolx to mark states with
                                 * additional information
                                 */
typedef posint *trans;
typedef trans  *ddelta;
/* det.transformation is a pointer to an array 
 *   (indexed with letter numbers from 1 to sno) 
 * and each element is a pointer to an array 
 *   (indexed with state numbers from 0 to qno) of state numbers. 
 * The transformation is (letter,state) to state.
 */

typedef struct dfauto
{ posint qno;			/* maximal state number, minimal state number is 0 */
  posint init;			/* initial state */
  posint sno;			/* size of alphabet */
  mrkfin final;			/* marks for final states */
  ddelta delta;			/* det. transition function */
  boole  minimal;		/* true iff automaton is minimal */
} *dfa;


/* types for the (epsilon-) nondeterministic finite automaton */

typedef char*        qlist;      /* bitmap representing a set of states */
typedef qlist*       ntrans;     /* transformation of a letter in a (e)nfa */
typedef ntrans*      ndelta;     /* nondet. transformation 
                                  * first  index   letter
                                  * second index   state
                                  * ndelta[i][j] is a bitmap 
                                  * representing all states
                                  * rechable from state j via letter i
                                  */

typedef struct nfauto
{ posint qno;			/* number of states */
  posint sno;			/* size of alphabet */
  mrkfin   infin;		/* marks for final and init. states */
  ndelta   delta;		/* ndet. transition relation */
  boole    minimal;		/* true iff automaton is minimal */
  boole    is_eps;		/* true iff epsilon-moves are present, 
				 * epsilon is letter 0 
				 */
} *nfa;

/* types for regular expressions */

typedef  struct rex
{ posint sno;			/* size of alphabet */
  string rex;			/* reg. expr. in infix, with abbreviations */
  string exprex;		/* reg. expr. in postfix, abbreviations expanded */
  string abbr[NABBR];		/* abbreviations in infix */
  posint abbl[NABBR];		/* length of abbr */
  int useda;			/* number of used abbr */
  posint rexl,erexl;		/* length of expressions */
  boole  grex;			/* TRUE iff generalized rex */
} *regex;

typedef  struct sexp 
{ array ulength,vlength,w0length,w1length; /* 0 .. mno-1 */
  arrayofarray u,v,w0,w1;	/* 0 .. mno-1 */
  /* length of u[x] :   2*ulength[x]   element,generator
   *           v[x] :   2*vlength[x]   generator,element
   *          w0[x] :   3*w0length[x]  generator,element,generator
   *          w1[x] :   1*w1length[x]     
   */
  b_array computed;		/* 0 .. mno-1 */
} *starfexp; 

/* types for monoids and monoid-structure */

typedef  struct dclass
{ posint    org;		/* element in the upper left hclass */
  posint    rno,lno;		/* #r-classes,#lclasses */
  posint    hsize;		/* #elements in a h-class */
  array     lrep,rrep;		/* array of prefixes(suffixes)
				 * of representatives of
				 * each r(l)-class (0..r(l)no-1) 
				 */
  array     hclass;		/* one full h-class (0..hsize-1) */
  boole     regular;		/* true iff regular d-class */
  posint    rang;		/* rang of d-class */
  posint    maxlen;		/* max. length of an element */
} *d_class;

typedef d_class *darray;

typedef struct d_struct
{ darray dclassarray;		/* (0..dno-1) */
  posint dno;			/* #dclasses */
  posint lno;			/* #lclasses */
  posint rno;			/* #rclasses */
} *dstruct;

typedef struct r_struct
{ posint    rno;		/* no of relations */
  array     lside,rside;	/* leftside(number of a monoidelement)
				 * rightside(number of a generator) 
				 * (0 .. rno-1)
				 */
} *rstruct;

typedef struct  mono
{ posint  qno,sno,mno;		/* #states,#letters,#elements */
  posint  gno;			/* no of generators */
  array   generator;		/* list of generators (0..gno)*/
  array   let2gen;		/* let2gen[i]=j iff 
				 * j is the least letter with the 
				 * transformation of generator i 
				 */
  arrayofarray no2trans;	/* transf. of elements(0..mno-1,0..qno) */
  arrayofarray gensucc;		/* gensucc[i][0]= predecessor of i      */
  /* gensucc[i][gen]=i*gen (0..mno-1,0..gno) */
  array     lastletter;		/* last letter of a shortest 
				 * representation of a element (0..mno-1) 
				 */
  array     no2length;		/* length of the representation (0..mno-1) */
  posint    zero;		/* number of the zero if one exists
				 * stamon->mno if no zero exists 
				 */
  boole     mequals;		/* syn. monoid = syn. semigroup */
  boole     dclassiscomputed;	/* d-classes are computed 
				 * no2rang is computed 
				 */
  dstruct   ds;			/* dclass structure if computed */
  boole     relationcomputed;
  rstruct   rs;			/* relation if computed */
  array     word;		/* free space used for the computation 
				 * of a representative with
				 * lastletter and gensucc[0] 
				 */
  string    repr;		/* free space for display of repr. */
} *monoid;


/* language */

typedef struct regset 
{ char      name[NLEN+1];	/* name of language */
  char      ref[REFLEN+1];	/* reference */
  boole     saved;		/* TRUE iff file exists */
  int       inputis;		/* 0 regex
				 * 1 dfa
				 * 2 nfa
				 * 3 eps-nfa
				 * 10 org. regex
				 * 11 org. dfa
				 * 12 org. nfa
				 * 13 org. eps-nfa
				 */
  posint    sizeal;		/* size of alphabet */
  regex     lrex;		/*  reg. expr. */  
  regex     orex;		/* original reg. expr. if inputis == 10 */
  boole     brex;		/* false iff lrex == nil */
  starfexp  sexp;		/* starfree expr. */
  boole     stariscomp;		/* starfree expression is computed */
  dfa       ldfa;		/* pt to a dfa */
  dfa       odfa;		/* original dfa if inputis == 11 */
  boole     bdfa;		/* false iff ldfa == nil */
  nfa       lnfa;		/* nfa */
  nfa       onfa;		/* original nfa if inputis == 12 */
  boole     bnfa;		/* false iff lnfa == nil */
  nfa       oenfa;		/* original enfa if inputis == 13 */
  monoid    lmon;		/* monoid */
  boole     bmon;		/* false iff lmon == nil */
  boolx     folu,
            empty,
            full,
            starfree,
            localtest,
            definite,
            revdefinite,
            gendefinite,
            dotdepth1,
            piecetest,
            nilpotent,
  /* --- Bordeaux : 1995 : start --- */      
            code,
            stack_code,
            prefix_code,
            suffix_code,
            bifix_code,
            p_infix_code,
            s_infix_code,
            infix_code,
            outfix_code,
            hypercode,
            right_semaphore_code,
            left_semaphore_code,
            uniform_code,
            full_uniform_code;
  char     *word_not_code;      /* word with two factorisation if not a code */
  /* --- Bordeaux : 1995 : end --- */      
  posint    localdegree;	/* localdegree 
				 * if degree is unknown lmon->mno 
				 */
} *language;

typedef struct tree_element
{
  char op;			/* Operand */
  string expr;
  int sons_no;			/* Number of sons */
  struct tree_element **son;	/* Pointerarray to sons */
  struct tree_element *father;	/* Pointer to father */
  posint son_passed,		/* Last son, that was already visited */
  q_no,				/* Number of states of automaton */
  init;				/* Initial state of automaton */
  boole is_reg_ab;		/* TRUE <=> no insecch or complch above node */
  dfa da;			/* Pointer to dfa */
  nfa na;			/* Pointer to nfa */
} *t_elem;

typedef struct str_rule
{
  string   leftside;
  string   rightside;
} *RULE;

typedef struct str_idgroups
{
  string  FirstSyms;		/* First symbols of the identifier(-group) */
  boole FollByNum;		/* TRUE iff identifier of the group consist of 
				   FirstSyms+ number */
  char  TypeSymbol;		/* Symbol representing the (group of) identifier(s) 
				   in the grammar */
} IdGroup;


typedef struct str_grammar
{
  posint         ruleno;
  char           startsymbol;
  RULE           *rules;
  IdGroup        *fident_list;
  arrayofstring  ident_list;
  arrayofstring  operator_list;
  arrayofstring  prefix_list;
  posint         maxid;
  posint         maxidlen;
  posint         maxop;
  posint         maxpref;
  arrayofb_array isinrule;
} *GRAMMAR;

typedef struct part {		/* part of a reg. expr. */
  int lastpos;			/* 0..lastpos-1 in info used */
  struct part *prev,*next;
  char *info;			/* string of information */
} *strstack;

typedef struct str_parse_result
{
  t_elem   tree;		/* Root of syntax-tree regular exp. */
  string   error_message;	/* Error message iff tree == NULL */
  posint   error_pos;		/* Position of error */
  posint   error_line;		/* Line in which error ocurred */
} *PARSE_RESULT;

typedef struct str_tbuf_elem
{
  char     type;
  string   symbols;
  posint   position;
  posint   line;
  t_elem   node;
} *tbuf_elem;

typedef struct str_symbol_stack
{
  char     operator;
  posint   position;
  posint   line;
  struct   str_symbol_stack *prev;
  struct   str_symbol_stack *next;
} *symbol_stack;

typedef struct str_sentence_stack
{
  char     operator;
  string   symbols;
  posint   position;
  posint   line;
  struct   str_sentence_stack *prev;
  struct   str_sentence_stack *next;
  t_elem   node;
} *sentence_stack;


typedef struct str_reduce_result
{
  string rule;
  posint synchr;
  sentence_stack s_elem;
  posint errorpos;
} *REDUCE_RESULT;

typedef struct str_parse_input
{
  string op_list;		/* List of operators */
  IdGroup *fid_list;		/* List of identifiers */
  string id_list;		/* List of identifiers */
  posint maxid;
  arrayofstring  rules;		/* Stringarray with the rules of the grammar */
  posint   maxrule;		/* Number of rules in the grammar */
  char  startsymbol;		/* Startsymbol of the grammar */
  arrayofarray precedence;	/* Precedence relation */
  strstack  buf;		/* Input, that shall be parsed */
  array index;			/* Mapping {operators}->|{operators}| */
} *PARSE_INPUT;
