/*
   Bordeaux : 1995
   */

#include "ext.h"

#include "win_una2.h"
#include "fa_utils.h"
#include "adjust.h"



dfa win_una2_root(dfa X_dfa, nfa X_nfa)
/* (X is the language of DFA X_dfa)
   Root(X) = ( X - {epsilon} ) - ( X - {epsilon} )^2 */
{
  dfa result_dfa;

  if( fa_utils_test_if_eps_in_dfa(X_dfa) == FALSE )
    {
      nfa X1_nfa;
      dfa X1_dfa;

      X1_nfa     = concatfa(X_nfa, X_nfa);
      X1_dfa     = nfa2dfa(X1_nfa);
      result_dfa = adjust_minus_dfa(X_dfa, X1_dfa);
      
      freedfa(X1_dfa);
      freenfa(X1_nfa);
    }
  else
    {
      dfa X1_dfa;
      dfa X2_dfa;
      dfa eps_dfa;
      
      eps_dfa    = fa_utils_eps_dfa(X_dfa->sno);
      /* language with only the empty word */
      X1_dfa     = adjust_minus_dfa(X_dfa, eps_dfa);
      X2_dfa     = adjust_concat_dfa(X1_dfa, X1_dfa);
      result_dfa = adjust_minus_dfa(X1_dfa, X2_dfa);
      
      fa_utils_free_temporary_fa(FALSE, 3, eps_dfa, X1_dfa, X2_dfa);
    }
      
  return result_dfa;
}




dfa win_una2_free_hull(dfa X_dfa, nfa X_nfa)
/* Algorithm : (X is the language of DFA X_dfa)
   if X in A+ :
     X_0 = X*
     X_(i+1) = [(X_i)(-1)X_i Inter X_i(X_i)(-1)]*
     free_hull(X) = Union X_i   
                    i>=0
   as the set of the X_i is finite, free_hull is computable
  */
{
  if(fa_utils_test_eps_dfa(X_dfa) == TRUE || 
     fa_utils_test_empty_dfa(X_dfa) == TRUE )
    return fa_utils_eps_dfa(X_dfa->sno);
  else
    {
      nfa tmp_nfa;
      dfa result_dfa;
      dfa empty_dfa;
      dfa X_i_dfa;
      dfa X_j_dfa;  /* j := i - 1 */
      
      empty_dfa  = fa_utils_empty_dfa(X_dfa->sno);  
      /* language with only the empty word */
      X_j_dfa    = fa_utils_duplicate_dfa(empty_dfa);
      result_dfa = fa_utils_duplicate_dfa(empty_dfa);
      tmp_nfa    = adjust_star_nfa(X_nfa);
      X_i_dfa    = nfa2dfa(tmp_nfa);
      freenfa(tmp_nfa);
      freedfa(empty_dfa);
      
      while( adjust_inclusion_dfa(X_i_dfa, result_dfa) == FALSE )
	{
	  dfa tmp1_dfa;
	  dfa tmp2_dfa;
	  dfa tmp3_dfa;
	  
	  result_dfa   = dfamdfa(adjust_union_dfa(result_dfa, X_i_dfa), TRUE);
	  X_j_dfa      = fa_utils_duplicate_dfa(X_i_dfa);
	  
	  tmp1_dfa   = rightquot(X_j_dfa, X_j_dfa);
	  tmp2_dfa   = adjust_leftquot_dfa(X_j_dfa, X_j_dfa);
	  tmp3_dfa   = dfamdfa(adjust_insec_dfa(tmp1_dfa, tmp2_dfa), TRUE);
	  
	  X_i_dfa = dfamdfa(adjust_star_dfa(tmp3_dfa), TRUE);
	  
	  fa_utils_free_temporary_fa(FALSE, 3, tmp1_dfa, tmp2_dfa, tmp3_dfa);
	}
      
      fa_utils_free_temporary_fa(FALSE, 2, X_i_dfa, X_j_dfa);
      
      return result_dfa;
    }
}




dfa win_una2_stack(dfa X_dfa)
/* Algorithm : (X is the language of DFA X_dfa)
   X_0 = X
   X_(i+1) = [X_i]*.([X_i]*)(-1)
   X# = Union X_i
        i>=0
   as  the set of X_i is finite, the satck_language is computable
   */
{
  if ( fa_utils_test_eps_dfa(X_dfa) == TRUE || fa_utils_test_empty_dfa(X_dfa) == TRUE )
    return fa_utils_eps_dfa(X_dfa->sno);
  else
    {
      dfa result_dfa;
      dfa empty_dfa;
      dfa X_j_dfa;
      dfa X_i_dfa; /* j := i - 1 */
      
      empty_dfa  = fa_utils_empty_dfa(X_dfa->sno);  
      /* language with only the empty word */
      result_dfa = fa_utils_duplicate_dfa(empty_dfa);
      X_j_dfa    = fa_utils_duplicate_dfa(empty_dfa);
      X_i_dfa    = fa_utils_duplicate_dfa(X_dfa);
      
      while ( adjust_inclusion_dfa(X_i_dfa, X_j_dfa) == FALSE )
	{
	  dfa star_X_i_dfa;
	  dfa tmp_dfa;
	  	  
	  tmp_dfa      = result_dfa;
	  result_dfa   = dfamdfa(adjust_union_dfa(tmp_dfa, X_i_dfa), TRUE);
	  star_X_i_dfa = adjust_star_dfa(X_i_dfa);
	  X_j_dfa      = X_i_dfa;    
	  X_i_dfa      = dfamdfa(rightquot(star_X_i_dfa, star_X_i_dfa), TRUE);
	  
	  fa_utils_free_temporary_fa(FALSE, 2, star_X_i_dfa, tmp_dfa);
	}
      
      fa_utils_free_temporary_fa(FALSE, 3, empty_dfa, X_i_dfa, X_j_dfa);
      
      return result_dfa;
    }
}









